"""
Generate and publish views for publicly available tables.

Generate view definitions for queries that are written to the
public data project and execute them. Views are published to
an internal project so that data is also accessible in private
datasets.
"""

from argparse import ArgumentParser

from google.cloud import bigquery

from ..config import ConfigLoader
from ..util import standard_args
from ..util.bigquery_tables import get_tables_matching_patterns

DEFAULT_PATTERN = "mozilla-public-data:*.*"


parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "--target-project",
    default=ConfigLoader.get("default", "project", fallback="moz-fx-data-shared-prod"),
    help="Create views in the target project",
)
parser.add_argument(
    "patterns",
    metavar="[project:]dataset[.table]",
    default=[DEFAULT_PATTERN],
    nargs="*",
    help="Table that should have a latest-version view, may use shell-style wildcards,"
    f" defaults to: {DEFAULT_PATTERN}",
)
standard_args.add_dry_run(parser)


def generate_and_publish_views(client, tables, target_project, dry_run):
    """Generate view definitions for public data tables and executes them."""
    for public_table in tables:
        project, dataset, table_name = public_table.split(".")
        full_view_id = f"{target_project}.{dataset}.{table_name}"

        view_sql = f"""CREATE OR REPLACE VIEW
            `{full_view_id}`
        AS SELECT * FROM `{public_table}`
        """

        job_config = bigquery.QueryJobConfig(use_legacy_sql=False, dry_run=dry_run)
        client.query(view_sql, job_config)


def main():
    """Publish public data views."""
    args = parser.parse_args()

    client = bigquery.Client(args.target_project)
    tables = get_tables_matching_patterns(client, args.patterns)
    generate_and_publish_views(client, tables, args.target_project, args.dry_run)


if __name__ == "__main__":
    main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, you can obtain one at http://mozilla.org/MPL/2.0/.
"""Utilities."""

import codecs
import json
import os
import os.path
import pprint
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import date, datetime
from decimal import Decimal
from io import BytesIO, TextIOWrapper
from typing import Any, Callable, Generator, List, Optional, Tuple, Union

import yaml
from google.api_core.exceptions import BadRequest, NotFound
from google.cloud import bigquery

QueryParameter = Union[
    bigquery.ArrayQueryParameter,
    bigquery.ScalarQueryParameter,
    bigquery.StructQueryParameter,
]

TABLE_EXTENSIONS = {
    "ndjson": bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
    "csv": bigquery.SourceFormat.CSV,
    "backup_info": bigquery.SourceFormat.DATASTORE_BACKUP,
    "export_metadata": bigquery.SourceFormat.DATASTORE_BACKUP,
    "avro": bigquery.SourceFormat.AVRO,
    "parquet": bigquery.SourceFormat.PARQUET,
    "orc": bigquery.SourceFormat.ORC,
}


@dataclass
class Table:
    """Define info needed to create a table for a generated test."""

    name: str
    source_format: str
    # a tuple means read via `load(*source_path)` and format as source_format
    # a string means source_path is already in source_format
    source_path: Union[str, Tuple[str, str]]
    # post_init fields
    schema: Optional[List[bigquery.SchemaField]] = None

    def __post_init__(self):
        """Fill in calculated fields if not provided."""
        if self.schema is None:
            if isinstance(self.source_path, str):
                resource_dir, resource = os.path.split(self.source_path)
                full_name, _ = resource.rsplit(".", 1)
            else:
                resource_dir, full_name = self.source_path
            try:
                table_dir, _ = os.path.split(resource_dir)
                self.schema = [
                    bigquery.SchemaField.from_api_repr(field)
                    for field in load(table_dir, f"{full_name}.schema")
                ]
            except FileNotFoundError:
                pass


class NDJsonDecodeError(Exception):
    """ndjson decode error."""

    pass


class JsonDecodeError(Exception):
    """JSON decode error."""

    pass


@contextmanager
def dataset(bq: bigquery.Client, dataset_id: str):
    """Context manager for creating and deleting the BigQuery dataset for a test."""
    try:
        result = bq.get_dataset(dataset_id)
    except NotFound:
        result = bq.create_dataset(dataset_id)
    try:
        yield result.reference
    finally:
        bq.delete_dataset(dataset_id, delete_contents=True)


def default_encoding(obj):
    """Add custom logic for serializing rows into JSON for BigQuery."""
    if isinstance(obj, date) or isinstance(obj, datetime):
        return obj.isoformat()
    return obj


def load_table(bq: bigquery.Client, dataset: bigquery.Dataset, table: Table):
    """Load table for a test."""
    destination = dataset.table(table.name)
    job_config = bigquery.LoadJobConfig(
        source_format=table.source_format,
        write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
    )

    if table.schema is None:
        # autodetect schema if not provided
        job_config.autodetect = True
    else:
        job_config.schema = table.schema
        # look for time_partitioning_field in provided schema
        for field in job_config.schema:
            if field.description == "time_partitioning_field":
                job_config.time_partitioning = bigquery.TimePartitioning(
                    field=field.name
                )
                break  # stop because there can only be one time partitioning field

    if isinstance(table.source_path, str):
        with open(table.source_path, "rb") as file_obj:
            job = bq.load_table_from_file(file_obj, destination, job_config=job_config)
    else:
        mem_file = BytesIO()
        for row in load(*table.source_path):
            mem_file.write(json.dumps(row, default=default_encoding).encode() + b"\n")
        mem_file.seek(0)
        job = bq.load_table_from_file(mem_file, destination, job_config=job_config)

    try:
        job.result()
    except BadRequest:
        # print the first 5 rows for debugging errors
        for row in job.errors[:5]:
            print(row)
        raise


def load_view(
    bq: bigquery.Client, dataset: bigquery.Dataset, view_name: str, view_query: str
):
    """Load view for a test."""
    view = bigquery.Table(dataset.table(view_name))
    view.view_query = view_query.format(
        project=dataset.project, dataset=dataset.dataset_id
    )
    bq.create_table(view)


def read(*paths: str, decoder: Optional[Callable] = None, **kwargs):
    """Read a file and apply decoder if provided."""
    with open(os.path.join(*paths), **kwargs) as f:
        return decoder(f) if decoder else f.read()


def ndjson_load(file_obj: TextIOWrapper) -> List[Any]:
    """Decode newline delimited json from file_obj."""
    res = []
    for i, line in enumerate(file_obj):
        try:
            res.append(json.loads(line))
        except json.JSONDecodeError as e:
            raise NDJsonDecodeError(
                f"Line {i+1} column {e.colno} of file {file_obj.name}, {e.msg}"
            )

    return res


def json_load(file_obj: TextIOWrapper) -> List[Any]:
    """Decode json from file_obj."""
    try:
        return json.load(file_obj)
    except json.JSONDecodeError as e:
        raise JsonDecodeError(
            f"Line {e.lineno} column {e.colno} of file {file_obj.name}, {e.msg}"
        )


def load(resource_dir: str, *basenames: str, **search: Optional[Callable]) -> Any:
    """Read the first matching file found in resource_dir.

    Calls read on paths under resource_dir with a name sans extension in
    basenames and an extension and decoder in search.

    :param resource_dir: directory to check for files
    :param basenames: file names to look for, without an extension
    :param search: mapping of file extension to decoder
    :return: first response from read() that doesn't raise FileNotFoundError
    :raises FileNotFoundError: when all matching files raise FileNotFoundError
    """
    search = search or {
        "yaml": yaml.full_load,
        "json": json_load,
        "ndjson": ndjson_load,
    }
    not_found: List[str] = []
    for basename in basenames:
        for ext, decoder in search.items():
            try:
                return read(resource_dir, f"{basename}.{ext}", decoder=decoder)
            except FileNotFoundError:
                not_found.append(f"{basename}.{ext}")
    raise FileNotFoundError(f"[Errno 2] No such files in '{resource_dir}': {not_found}")


def get_query_params(resource_dir: str) -> Generator[QueryParameter, None, None]:
    """Attempt to load the first query params found in resource_dir."""
    try:
        params = load(resource_dir, "query_params")
    except FileNotFoundError:
        params = []
    for param in params:
        if {"name", "type", "type_", "value"}.issuperset(param.keys()):
            # this is a scalar query param
            param["type_"] = param.pop("type", param.pop("type_", "STRING"))
            yield bigquery.ScalarQueryParameter(**param)
        else:
            # attempt to coerce to some type of query param
            try:
                yield bigquery.StructQueryParameter.from_api_repr(param)
            except KeyError:
                try:
                    yield bigquery.ArrayQueryParameter.from_api_repr(param)
                except KeyError:
                    # this is a different format for scalar param than above
                    yield bigquery.ScalarQueryParameter.from_api_repr(param)


def coerce_result(*elements: Any) -> Generator[Any, None, None]:
    """Recursively coerce elements to types available in json.

    Coerce date and datetime to string using isoformat.
    Coerce bigquery.Row to dict using comprehensions.
    Coerce bytes to base64 encoded strings.
    Omit dict keys named "generated_time".
    Omit columns with null results to simplify `expect` files.
    """
    for element in elements:
        if isinstance(element, (dict, bigquery.Row)):
            yield {
                key: (
                    list(coerce_result(*value))
                    if isinstance(value, list)
                    else next(coerce_result(value))
                )
                for key, value in element.items()
                # drop generated_time column
                if key not in ("generated_time",) and value is not None
            }
        elif isinstance(element, (date, datetime)):
            yield element.isoformat()
        elif isinstance(element, Decimal):
            yield str(element)
        elif isinstance(element, bytes):
            yield codecs.encode(element, "base64").decode().strip()
        else:
            yield element


def get_differences(expected, result, path="", sep="."):
    """Get the differences between two JSON-like python objects.

    For complicated objects, this is a big improvement over pytest -vv
    """
    differences = []

    if expected is not None and result is None:
        differences.append(("Expected exists but not Result", path))
    if expected is None and result is not None:
        differences.append(("Result exists but not Expected", path))
    if expected is None and result is None:
        return differences

    exp_is_dict, res_is_dict = isinstance(expected, dict), isinstance(result, dict)
    exp_is_list, res_is_list = isinstance(expected, list), isinstance(result, list)
    if exp_is_dict and not res_is_dict:
        differences.append(("Expected is dict but not Result", path))
    elif res_is_dict and not exp_is_dict:
        differences.append(("Result is dict but not Expected", path))
    elif not exp_is_dict and not res_is_dict:
        if exp_is_list and res_is_list:
            for i in range(max(len(expected), len(result))):
                if i >= len(result):
                    differences.append(
                        (f"Result missing element {expected[i]}", path + sep + str(i))
                    )
                elif i >= len(expected):
                    differences.append(
                        (
                            f"Result contains extra element {result[i]}",
                            path + sep + str(i),
                        )
                    )
                else:
                    differences += get_differences(
                        expected[i], result[i], path + sep + str(i)
                    )
        elif expected != result:
            differences.append((f"Expected={expected}, Result={result}", path))
    else:
        exp_keys, res_keys = set(expected.keys()), set(result.keys())
        in_exp_not_res, in_res_not_exp = exp_keys - res_keys, res_keys - exp_keys

        for k in in_exp_not_res:
            differences.append(("In Expected, not in Result", path + sep + k))
        for k in in_res_not_exp:
            differences.append(("In Result, not in Expected", path + sep + k))

        for k in exp_keys & res_keys:
            differences += get_differences(expected[k], result[k], path + sep + k)

    return differences


def print_and_test(expected, result):
    """Print objects and differences, then test equality."""
    pp = pprint.PrettyPrinter(indent=2)

    print("\nExpected:")
    pp.pprint(expected)

    print("\nActual:")
    pp.pprint(result)

    print("\nDifferences:")
    print("\n".join([" - ".join(v) for v in get_differences(expected, result)]))

    assert result == expected



import os
from textwrap import dedent
from unittest.mock import Mock, patch

import yaml
from click.testing import CliRunner

from bigquery_etl.cli.query import extract_and_run_temp_udfs, run


class TestRunQuery:
    def test_run_query(self, tmp_path):
        query_file_path = tmp_path / "sql" / "test" / "query_v1"
        os.makedirs(query_file_path)
        query_file = query_file_path / "query.sql"
        query_file.write_text("-- comment \n SELECT 1")

        metadata_conf = {
            "friendly_name": "test",
            "description": "test",
            "owners": ["test@example.org"],
        }

        metadata_file = query_file_path / "metadata.yaml"
        metadata_file.write_text(yaml.dump(metadata_conf))

        runner = CliRunner()
        with patch("subprocess.check_call") as mock_call:
            mock_call.return_value = True
            result = runner.invoke(
                run,
                [str(query_file), "--dataset_id=test", "--destination_table=query_v1"],
            )

            assert result.exit_code == 0

            assert mock_call.call_args.args == (
                [
                    "bq",
                    "query",
                    "--destination_table=query_v1",
                    "--dataset_id=test",
                ],
            )
            assert "stdin" in mock_call.call_args.kwargs

    def test_run_query_qualified_destination_table(self, tmp_path):
        query_file_path = tmp_path / "sql" / "test" / "query_v1"
        os.makedirs(query_file_path)
        query_file = query_file_path / "query.sql"
        query_file.write_text("-- comment \n SELECT 1")

        metadata_conf = {
            "friendly_name": "test",
            "description": "test",
            "owners": ["test@example.org"],
        }

        metadata_file = query_file_path / "metadata.yaml"
        metadata_file.write_text(yaml.dump(metadata_conf))

        runner = CliRunner()
        with patch("subprocess.check_call") as mock_call:
            mock_call.return_value = True
            result = runner.invoke(
                run,
                [
                    str(query_file),
                    "--dataset_id=test",
                    "--destination_table=mozdata:test.query_v1",
                ],
            )

            assert result.exit_code == 0

            assert mock_call.call_args.args == (
                [
                    "bq",
                    "query",
                    "--destination_table=mozdata:test.query_v1",
                    "--dataset_id=test",
                ],
            )
            assert "stdin" in mock_call.call_args.kwargs

    def test_run_query_public_project(self, tmp_path):
        query_file_path = tmp_path / "sql" / "test" / "query_v1"
        os.makedirs(query_file_path)
        query_file = query_file_path / "query.sql"
        query_file.write_text("SELECT 1")

        metadata_conf = {
            "friendly_name": "test",
            "description": "test",
            "owners": ["test@example.org"],
            "labels": {"public_bigquery": True, "review_bugs": [222222]},
        }

        metadata_file = query_file_path / "metadata.yaml"
        metadata_file.write_text(yaml.dump(metadata_conf))

        with patch("subprocess.check_call") as mock_call:
            mock_call.return_value = True
            runner = CliRunner()
            result = runner.invoke(
                run,
                [str(query_file), "--dataset_id=test", "--destination_table=query_v1"],
            )

            assert result.exit_code == 0

            assert mock_call.call_args.args == (
                [
                    "bq",
                    "query",
                    "--destination_table=mozilla-public-data:test.query_v1",
                    "--dataset_id=test",
                ],
            )

    def test_run_query_public_project_no_dataset(self, tmp_path):
        query_file_path = tmp_path / "sql" / "test" / "query_v1"
        os.makedirs(query_file_path)
        query_file = query_file_path / "query.sql"
        query_file.write_text("SELECT 1")

        metadata_conf = {
            "friendly_name": "test",
            "description": "test",
            "owners": ["test@example.org"],
            "labels": {"public_bigquery": True, "review_bugs": [222222]},
        }

        metadata_file = query_file_path / "metadata.yaml"
        metadata_file.write_text(yaml.dump(metadata_conf))
        runner = CliRunner()

        with patch("subprocess.check_call") as mock_call:
            mock_call.return_value = True
            result = runner.invoke(
                run, [str(query_file), "--destination_table=query_v1"]
            )
            assert result.exit_code == 1

            result = runner.invoke(run, [str(query_file), "--dataset_id=test"])
            assert result.exit_code == 1

    @patch("subprocess.check_call")
    @patch("google.cloud.bigquery.Client")
    def test_run_query_billing_project(
        self, mock_client, mock_subprocess_call, tmp_path
    ):
        query_file_path = (
            tmp_path / "sql" / "moz-fx-data-shared-prod" / "dataset_1" / "query_v1"
        )
        os.makedirs(query_file_path)
        query_file = query_file_path / "query.sql"
        query_file.write_text("SELECT 1")

        runner = CliRunner()

        # bigquery.client().query().session_info.session_id = ...
        mock_query_call = Mock()
        mock_query_call.return_value.session_info.session_id = "1234567890"
        mock_client.return_value.query = mock_query_call

        mock_subprocess_call.return_value = 1
        result = runner.invoke(
            run,
            [
                str(query_file),
                "--billing-project=project-2",
                "--project-id=moz-fx-data-shared-prod",
            ],
        )
        assert result.exit_code == 0

        assert mock_query_call.call_count == 1
        query_text, query_job_config = mock_query_call.call_args.args
        assert (
            query_text
            == "SET @@dataset_project_id = 'moz-fx-data-shared-prod';\nSET @@dataset_id = 'dataset_1'"
        )
        assert query_job_config.create_session is True

        assert mock_subprocess_call.call_args.args == (
            [
                "bq",
                "query",
                "--project_id=project-2",
                "--session_id=1234567890",
            ],
        )

    @patch("subprocess.check_call")
    @patch("google.cloud.bigquery.Client")
    def test_run_query_billing_project_temp_udf(
        self, mock_client, mock_subprocess_call, tmp_path
    ):
        query_file_path = (
            tmp_path / "sql" / "moz-fx-data-shared-prod" / "dataset_1" / "query_v1"
        )
        os.makedirs(query_file_path)
        query_file = query_file_path / "query.sql"

        temp_udf_sql = "CREATE TEMP FUNCTION fn(param INT) AS (1);"
        query_file.write_text(f"{temp_udf_sql} SELECT 1")

        runner = CliRunner()

        # bigquery.client().query().session_info.session_id = ...
        mock_query_call = Mock()
        mock_query_call.return_value.session_info.session_id = "1234567890"
        mock_client.return_value.query = mock_query_call

        mock_query_and_wait_call = Mock()
        session_id = "1234567890"
        mock_client.return_value.query_and_wait = mock_query_and_wait_call

        mock_subprocess_call.return_value = 1
        result = runner.invoke(
            run,
            [
                str(query_file),
                "--billing-project=project-2",
                "--project-id=moz-fx-data-shared-prod",
                "--destination-table=proj.dataset.table",
            ],
        )
        assert result.exit_code == 0

        assert mock_query_call.call_count == 1

        # udf define query
        assert mock_query_and_wait_call.call_count == 1
        query_text = mock_query_and_wait_call.call_args[0][0]
        query_job_config = mock_query_and_wait_call.call_args[1]["job_config"]
        assert query_text == temp_udf_sql
        assert query_job_config.connection_properties[0].value == session_id

    @patch("google.cloud.bigquery.Client")
    def test_extract_and_run_temp_udfs(self, mock_client):
        mock_client.return_value = mock_client

        udf_sql = dedent(
            """
            CREATE TEMP FUNCTION f1(arr ARRAY<INT64>) AS (
              (SELECT SUM(a) FROM UNNEST(arr) AS a)
            );

            CREATE  -- inline comment
              TEMPORARY
                FUNCTION f2() AS (
              (SELECT 1)
            );

            CREATE TEMP FUNCTION f3() RETURNS INT64 AS (1);

            -- javascript
            CREATE TEMP FUNCTION f4(input JSON)
            RETURNS JSON
            DETERMINISTIC
            LANGUAGE js
            AS
            \"\"\"
                return "abc";
            \"\"\";

            CREATE TEMP FUNCTION f5()
            RETURNS STRING LANGUAGE js
            AS "return 'abc'";
            """
        )
        query_sql = dedent(
            """
            WITH abc AS (
              SELECT * FROM UNNEST([1, 2, 3]) AS n
            )
            SELECT "CREATE TEMP FUNCTION f3() AS (1);", * FROM abc
            """
        )
        sql = f"{udf_sql}{query_sql}"

        updated_query = extract_and_run_temp_udfs(
            sql, project_id="project-1", session_id="123"
        )

        assert updated_query == query_sql.strip()

        mock_client.assert_called_once_with(project="project-1")

        assert mock_client.query_and_wait.call_count == 1

        # remove empty lines for comparison
        udf_sql = "\n".join([line for line in udf_sql.split("\n") if line.strip()])
        assert mock_client.query_and_wait.call_args[0][0] == udf_sql



import os
from pathlib import Path

import pytest

from bigquery_etl.metadata.parse_metadata import Metadata
from bigquery_etl.query_scheduling.dag import DagParseException, InvalidDag
from bigquery_etl.query_scheduling.dag_collection import DagCollection
from bigquery_etl.query_scheduling.task import Task, TaskRef

TEST_DIR = Path(__file__).parent.parent


class TestDagCollection:
    default_args = {
        "start_date": "2020-05-15",
        "owner": "test@example.org",
        "email": ["test@example.org"],
        "depends_on_past": False,
    }

    def test_dags_from_file(self):
        dags_file = TEST_DIR / "data" / "dags.yaml"
        dags = DagCollection.from_file(dags_file)

        assert len(dags.dags) == 2
        assert dags.dag_by_name("not existing") is None
        assert dags.dag_by_name("bqetl_events") is not None
        assert dags.dag_by_name("bqetl_core") is not None

        events_dag = dags.dag_by_name("bqetl_events")
        assert len(events_dag.tasks) == 0

        core_dag = dags.dag_by_name("bqetl_core")
        assert len(core_dag.tasks) == 0

    def test_dags_from_empty_file(self, tmp_path):
        dags_file = tmp_path / "dags.yaml"
        dags_file.write_text("")
        dags = DagCollection.from_file(dags_file)

        assert len(dags.dags) == 0

    def test_dags_from_dict(self):
        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag1": {
                    "schedule_interval": "daily",
                    "default_args": self.default_args,
                },
                "bqetl_test_dag2": {
                    "schedule_interval": "daily",
                    "default_args": self.default_args,
                },
            }
        )

        assert len(dags.dags) == 2
        assert dags.dag_by_name("bqetl_test_dag1") is not None
        assert dags.dag_by_name("bqetl_test_dag2") is not None

        dag1 = dags.dag_by_name("bqetl_test_dag1")
        assert len(dag1.tasks) == 0
        assert dag1.schedule_interval == "daily"
        assert dag1.default_args.owner == "test@example.org"

        dag2 = dags.dag_by_name("bqetl_test_dag2")
        assert len(dag2.tasks) == 0
        assert dag2.schedule_interval == "daily"

    def test_dags_from_empty_dict(self):
        dags = DagCollection.from_dict({})
        assert len(dags.dags) == 0

    def test_dags_from_invalid_dict(self):
        with pytest.raises(DagParseException):
            DagCollection.from_dict({"foo": "bar"})

    def test_dag_by_name(self):
        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag1": {
                    "schedule_interval": "daily",
                    "default_args": self.default_args,
                }
            }
        )

        assert dags.dag_by_name("bqetl_test_dag1") is not None
        assert dags.dag_by_name("bqetl_test_dag1").name == "bqetl_test_dag1"
        assert dags.dag_by_name("non_existing") is None

    def test_task_for_table(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        metadata = Metadata(
            "test",
            "test",
            ["test@example.org"],
            {},
            {
                "dag_name": "bqetl_test_dag",
                "depends_on_past": True,
                "secrets": [
                    {
                        "deploy_target": "SECRET",
                        "key": "some_secret_stored",
                        "deploy_type": "env",
                        "secret": "gke-secrets",
                    },
                    {
                        "deploy_target": "SECRET2",
                        "key": "some_secret_stored_2",
                        "deploy_type": "env",
                        "secret": "gke-secrets",
                    },
                ],
            },
        )

        tasks = [Task.of_query(query_file, metadata)]

        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": self.default_args,
                }
            }
        ).with_tasks(tasks)

        task = dags.task_for_table(
            "moz-fx-data-test-project", "test", "incremental_query_v1"
        )

        assert task
        assert task.dag_name == "bqetl_test_dag"
        assert len(task.secrets) == 2

    def test_task_for_non_existing_table(self):
        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": self.default_args,
                }
            }
        ).with_tasks([])

        assert (
            dags.task_for_table(
                "moz-fx-data-test-project", "test", "non_existing_table"
            )
            is None
        )

    def test_dags_with_tasks(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        metadata = Metadata(
            "test",
            "test",
            ["test@example.org"],
            {},
            {"dag_name": "bqetl_test_dag", "depends_on_past": True},
        )

        tasks = [Task.of_query(query_file, metadata)]

        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": self.default_args,
                }
            }
        ).with_tasks(tasks)

        assert len(dags.dags) == 1

        dag = dags.dag_by_name("bqetl_test_dag")
        assert len(dag.tasks) == 1
        assert dag.tasks[0].dag_name == "bqetl_test_dag"

    def test_dags_with_invalid_tasks(self):
        with pytest.raises(InvalidDag):
            query_file = (
                TEST_DIR
                / "data"
                / "moz-fx-data-test-project"
                / "test_sql"
                / "test"
                / "incremental_query_v1"
                / "query.sql"
            )

            metadata = Metadata(
                "test",
                "test",
                ["test@example.org"],
                {},
                {
                    "dag_name": "bqetl_non_exisiting_dag",
                    "depends_on_past": True,
                    "param": "test_param",
                },
            )

            tasks = [Task.of_query(query_file, metadata)]

            DagCollection.from_dict(
                {
                    "bqetl_test_dag": {
                        "schedule_interval": "daily",
                        "default_args": self.default_args,
                    }
                }
            ).with_tasks(tasks)

    def test_to_airflow(self, tmp_path):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "non_incremental_query_v1"
            / "query.sql"
        )

        metadata = Metadata(
            "test",
            "test",
            ["test@example.com"],
            {},
            {
                "dag_name": "bqetl_test_dag",
                "depends_on_past": True,
                "param": "test_param",
                "arguments": ["--append_table"],
                "depends_on_fivetran": [
                    {"task_id": "fivetran_import_1"},
                    {"task_id": "fivetran_import_2"},
                ],
            },
        )

        tasks = [Task.of_query(query_file, metadata)]

        default_args = {
            "depends_on_past": False,
            "owner": "test@example.org",
            "email": ["test@example.org"],
            "start_date": "2020-01-01",
            "retry_delay": "1h",
        }
        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": default_args,
                }
            }
        ).with_tasks(tasks)

        dags.to_airflow_dags(tmp_path)
        result = (tmp_path / "bqetl_test_dag.py").read_text().strip()
        expected = (TEST_DIR / "data" / "dags" / "simple_test_dag").read_text().strip()
        assert result == expected

    def test_python_script_to_airflow(self, tmp_path):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "python_script_query_v1"
            / "query.py"
        )

        metadata = Metadata(
            "test",
            "test",
            ["test@example.com"],
            {},
            {
                "dag_name": "bqetl_test_dag",
                "depends_on_past": True,
                "arguments": ["--date", "{{ds}}"],
                "task_group": "test_group",
            },
        )

        tasks = [
            Task.of_python_script(query_file, metadata),
            Task.of_bigeye_check(query_file, metadata),
        ]

        default_args = {
            "depends_on_past": False,
            "owner": "test@example.org",
            "email": ["test@example.org"],
            "start_date": "2020-01-01",
            "retry_delay": "1h",
        }
        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": default_args,
                }
            }
        ).with_tasks(tasks)

        dags.to_airflow_dags(tmp_path)
        result = (tmp_path / "bqetl_test_dag.py").read_text().strip()
        expected = (
            (TEST_DIR / "data" / "dags" / "python_script_test_dag").read_text().strip()
        )

        assert result == expected

    def test_to_airflow_with_upstream_dependencies(self, tmp_path):
        query_file_path = tmp_path / "test-project" / "test" / "query_v1"
        os.makedirs(query_file_path)

        query_file = query_file_path / "query.sql"
        query_file.write_text(
            "SELECT * FROM `test-project`.test.table1_v1 "
            "UNION ALL SELECT * FROM `test-project`.test.table2_v1 "
            "UNION ALL SELECT * FROM `test-project`.test.external_table_v1"
        )

        metadata = Metadata(
            "test",
            "test",
            ["test@example.org"],
            {},
            {
                "dag_name": "bqetl_test_dag",
                "default_args": {"owner": "test@example.org"},
            },
        )

        task = Task.of_query(query_file, metadata)

        table_task1 = Task.of_query(
            tmp_path / "test-project" / "test" / "table1_v1" / "query.sql",
            metadata,
        )

        os.makedirs(tmp_path / "test-project" / "test" / "table1_v1")
        query_file = tmp_path / "test-project" / "test" / "table1_v1" / "query.sql"
        query_file.write_text("SELECT 1")

        table_task2 = Task.of_query(
            tmp_path / "test-project" / "test" / "table2_v1" / "query.sql",
            metadata,
        )

        os.makedirs(tmp_path / "test-project" / "test" / "table2_v1")
        query_file = tmp_path / "test-project" / "test" / "table2_v1" / "query.sql"
        query_file.write_text("SELECT 2")

        metadata = Metadata(
            "test",
            "test",
            ["test@example.org"],
            {},
            {
                "dag_name": "bqetl_external_test_dag",
                "default_args": {"owner": "test@example.org"},
            },
        )

        external_table_task = Task.of_query(
            tmp_path / "test-project" / "test" / "external_table_v1" / "query.sql",
            metadata,
        )

        os.makedirs(tmp_path / "test-project" / "test" / "external_table_v1")
        query_file = (
            tmp_path / "test-project" / "test" / "external_table_v1" / "query.sql"
        )
        query_file.write_text("SELECT 3")

        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-05-25",
                    },
                },
                "bqetl_external_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-05-25",
                    },
                },
            }
        ).with_tasks([task, table_task1, table_task2, external_table_task])

        dags.to_airflow_dags(tmp_path)

        expected_dag_with_upstream_dependencies = (
            (TEST_DIR / "data" / "dags" / "test_dag_with_dependencies")
            .read_text()
            .strip()
        )
        expected_dag_external_dependency = (
            (TEST_DIR / "data" / "dags" / "test_dag_external_dependency")
            .read_text()
            .strip()
        )

        dag_with_upstream_dependencies = (
            (tmp_path / "bqetl_test_dag.py").read_text().strip()
        )
        dag_external_dependency = (
            (tmp_path / "bqetl_external_test_dag.py").read_text().strip()
        )

        assert dag_with_upstream_dependencies == expected_dag_with_upstream_dependencies
        assert dag_external_dependency == expected_dag_external_dependency

    def test_public_json_dag_to_airflow(self, tmp_path):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "non_incremental_query_v1"
            / "query.sql"
        )

        tasks = [Task.of_query(query_file)]

        default_args = {
            "depends_on_past": False,
            "owner": "test@example.org",
            "email": ["test@example.org"],
            "start_date": "2020-01-01",
            "retry_delay": "1h",
        }

        dags = DagCollection.from_dict(
            {
                "bqetl_public_data_json": {
                    "schedule_interval": "daily",
                    "default_args": default_args,
                },
                "bqetl_core": {
                    "schedule_interval": "daily",
                    "default_args": default_args,
                },
            }
        ).with_tasks(tasks)

        dags.to_airflow_dags(tmp_path)
        result = (tmp_path / "bqetl_public_data_json.py").read_text().strip()
        expected_dag = (
            (TEST_DIR / "data" / "dags" / "test_public_data_json_dag")
            .read_text()
            .strip()
        )

        assert result == expected_dag

    def test_to_airflow_duplicate_dependencies(self, tmp_path):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "non_incremental_query_v1"
            / "query.sql"
        )

        query_file2 = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "no_metadata_query_v1"
            / "query.sql"
        )

        metadata = Metadata(
            "test",
            "test",
            ["test@example.com"],
            {},
            {
                "dag_name": "bqetl_test_dag",
                "depends_on_past": True,
                "date_partition_offset": -2,
                "date_partition_parameter": "date",
                "depends_on": [{"dag_name": "external", "task_id": "task1"}],
            },
        )

        tasks = [
            Task.of_query(query_file, metadata),
            Task.of_query(query_file2, metadata),
        ]

        default_args = {
            "owner": "test@example.org",
            "start_date": "2020-01-01",
        }
        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": default_args,
                    "tags": ["repo/bigquery-etl"],
                }
            }
        ).with_tasks(tasks)

        dags.to_airflow_dags(tmp_path)
        result = (tmp_path / "bqetl_test_dag.py").read_text().strip()
        expected = (
            (TEST_DIR / "data" / "dags" / "test_dag_duplicate_dependencies")
            .read_text()
            .strip()
        )

        assert result == expected

    def test_to_airflow_with_check_upstream_dependencies(self, tmp_path):
        query_file_path = tmp_path / "test-project" / "test" / "query_v1"
        os.makedirs(query_file_path)

        query_file = query_file_path / "query.sql"
        query_file.write_text(
            "SELECT * FROM `test-project`.test.table1_v1 "
            "UNION ALL SELECT * FROM `test-project`.test.table2_v1 "
            "UNION ALL SELECT * FROM `test-project`.test.external_table_v1"
        )

        metadata = Metadata(
            "test",
            "test",
            ["test@example.org"],
            {},
            {
                "dag_name": "bqetl_test_dag",
                "default_args": {"owner": "test@example.org"},
            },
        )

        task = Task.of_query(query_file, metadata)

        table_task1 = Task.of_query(
            tmp_path / "test-project" / "test" / "table1_v1" / "query.sql",
            metadata,
        )
        table_task1_ref = TaskRef(
            dag_name=table_task1.dag_name, task_id=table_task1.task_name
        )

        os.makedirs(tmp_path / "test-project" / "test" / "table1_v1")
        query_file = tmp_path / "test-project" / "test" / "table1_v1" / "query.sql"
        query_file.write_text("SELECT 1")

        checks_task1 = Task.of_dq_check(
            tmp_path / "test-project" / "test" / "table1_v1" / "checks.sql",
            is_check_fail=True,
            metadata=metadata,
        )

        checks_task1.upstream_dependencies.append(table_task1_ref)

        check_file1 = tmp_path / "test-project" / "test" / "table1_v1" / "checks.sql"
        check_file1.write_text("SELECT TRUE")

        table_task2 = Task.of_query(
            tmp_path / "test-project" / "test" / "table2_v1" / "query.sql",
            metadata,
        )

        os.makedirs(tmp_path / "test-project" / "test" / "table2_v1")
        query_file = tmp_path / "test-project" / "test" / "table2_v1" / "query.sql"
        query_file.write_text("SELECT 2")

        metadata = Metadata(
            "test",
            "test",
            ["test@example.org"],
            {},
            {
                "dag_name": "bqetl_external_test_dag",
                "default_args": {"owner": "test@example.org"},
            },
        )

        external_table_task = Task.of_query(
            tmp_path / "test-project" / "test" / "external_table_v1" / "query.sql",
            metadata,
        )
        external_table_task_ref = TaskRef(
            dag_name=external_table_task.dag_name, task_id=external_table_task.task_name
        )

        os.makedirs(tmp_path / "test-project" / "test" / "external_table_v1")
        query_file = (
            tmp_path / "test-project" / "test" / "external_table_v1" / "query.sql"
        )
        query_file.write_text("SELECT 3")

        checks_task2 = Task.of_dq_check(
            tmp_path / "test-project" / "test" / "external_table_v1" / "checks.sql",
            is_check_fail=True,
            metadata=metadata,
        )
        checks_task2.upstream_dependencies.append(external_table_task_ref)

        check_file2 = (
            tmp_path / "test-project" / "test" / "external_table_v1" / "checks.sql"
        )
        check_file2.write_text("SELECT TRUE")

        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-05-25",
                    },
                },
                "bqetl_external_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-05-25",
                    },
                },
            }
        ).with_tasks(
            [
                task,
                table_task1,
                table_task2,
                external_table_task,
                checks_task1,
                checks_task2,
            ]
        )

        dags.to_airflow_dags(tmp_path)

        expected_dag_with_upstream_dependencies = (
            (TEST_DIR / "data" / "dags" / "test_dag_with_check_dependencies")
            .read_text()
            .strip()
        )
        expected_dag_external_dependency = (
            (TEST_DIR / "data" / "dags" / "test_dag_external_check_dependency")
            .read_text()
            .strip()
        )

        dag_with_upstream_dependencies = (
            (tmp_path / "bqetl_test_dag.py").read_text().strip()
        )

        dag_external_dependency = (
            (tmp_path / "bqetl_external_test_dag.py").read_text().strip()
        )

        assert dag_with_upstream_dependencies == expected_dag_with_upstream_dependencies
        assert dag_external_dependency == expected_dag_external_dependency

    def test_to_airflow_with_check_table_dependencies(self, tmp_path):
        metadata = Metadata(
            "test",
            "test",
            ["test@example.org"],
            {},
            {
                "dag_name": "bqetl_test_dag",
                "default_args": {"owner": "test@example.org"},
            },
        )

        table_task1 = Task.of_query(
            tmp_path / "test-project" / "test" / "table1_v1" / "query.sql",
            metadata,
        )
        table_task1_ref = TaskRef(
            dag_name=table_task1.dag_name, task_id=table_task1.task_name
        )

        os.makedirs(tmp_path / "test-project" / "test" / "table1_v1")
        query_file = tmp_path / "test-project" / "test" / "table1_v1" / "query.sql"
        query_file.write_text("SELECT 1")

        checks_task1 = Task.of_dq_check(
            tmp_path / "test-project" / "test" / "table1_v1" / "checks.sql",
            is_check_fail=True,
            metadata=metadata,
        )

        checks_task1.upstream_dependencies.append(table_task1_ref)

        check_file1 = tmp_path / "test-project" / "test" / "table1_v1" / "checks.sql"
        check_file1.write_text("SELECT * FROM `test-project`.test.table2_v1")

        table_task2 = Task.of_query(
            tmp_path / "test-project" / "test" / "table2_v1" / "query.sql",
            metadata,
        )
        table_task2_ref = TaskRef(
            dag_name=table_task2.dag_name, task_id=table_task2.task_name
        )

        checks_task1.upstream_dependencies.append(table_task2_ref)

        os.makedirs(tmp_path / "test-project" / "test" / "table2_v1")
        query_file = tmp_path / "test-project" / "test" / "table2_v1" / "query.sql"
        query_file.write_text("SELECT 2")

        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-05-25",
                    },
                }
            }
        ).with_tasks(
            [
                table_task1,
                table_task2,
                checks_task1,
            ]
        )

        dags.to_airflow_dags(tmp_path)

        expected_dag_with_upstream_dependencies = (
            (TEST_DIR / "data" / "dags" / "test_dag_with_check_table_dependencies")
            .read_text()
            .strip()
        )

        dag_with_upstream_dependencies = (
            (tmp_path / "bqetl_test_dag.py").read_text().strip()
        )

        assert dag_with_upstream_dependencies == expected_dag_with_upstream_dependencies

    def test_to_airflow_with_bigquery_table_sensors(self, tmp_path):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "non_incremental_query_v1"
            / "query.sql"
        )

        metadata = Metadata(
            "test",
            "test",
            owners=["test@example.com"],
            labels={},
            scheduling={
                "dag_name": "bqetl_test_dag",
                "depends_on_past": True,
                "param": "test_param",
                "arguments": ["--append_table"],
                "depends_on_tables_existing": [
                    {
                        "task_id": "wait_for_foo_bar_baz",
                        "table_id": "foo.bar.baz_{{ ds_nodash }}",
                        "poke_interval": "30m",
                        "timeout": "12h",
                        "retries": 1,
                        "retry_delay": "10m",
                    },
                ],
                "depends_on_table_partitions_existing": [
                    {
                        "task_id": "wait_for_foo_bar_baz_partition",
                        "table_id": "foo.bar.baz",
                        "partition_id": "{{ ds_nodash }}",
                        "poke_interval": "15m",
                        "timeout": "6h",
                        "retries": 3,
                        "retry_delay": "5m",
                    },
                ],
            },
        )

        tasks = [Task.of_query(query_file, metadata)]

        default_args = {
            "depends_on_past": False,
            "owner": "test@example.org",
            "email": ["test@example.org"],
            "start_date": "2020-01-01",
            "retry_delay": "1h",
        }
        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": default_args,
                }
            }
        ).with_tasks(tasks)

        dags.to_airflow_dags(tmp_path)
        result = (tmp_path / "bqetl_test_dag.py").read_text().strip()
        expected = (
            (TEST_DIR / "data" / "dags" / "test_dag_with_bigquery_table_sensors")
            .read_text()
            .strip()
        )
        assert result == expected

    def test_to_airflow_with_secrets(self, tmp_path):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        metadata = Metadata.from_file(
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "metadata.yaml"
        )

        tasks = [Task.of_query(query_file, metadata)]

        default_args = {
            "depends_on_past": False,
            "owner": "test@example.org",
            "email": ["test@example.org"],
            "start_date": "2020-01-01",
            "retry_delay": "1h",
        }
        dags = DagCollection.from_dict(
            {
                "bqetl_events": {
                    "schedule_interval": "daily",
                    "default_args": default_args,
                }
            }
        ).with_tasks(tasks)

        dags.to_airflow_dags(tmp_path)
        result = (tmp_path / "bqetl_events.py").read_text().strip()
        print(result)
        expected = (
            (TEST_DIR / "data" / "dags" / "test_dag_with_secrets").read_text().strip()
        )
        assert result == expected



import os
from pathlib import Path

import pytest
from click.testing import CliRunner

from bigquery_etl.cli.utils import is_valid_dir
from bigquery_etl.util.common import (
    extract_last_group_by_from_query,
    project_dirs,
    qualify_table_references_in_file,
    render,
)


class TestUtilCommon:
    @pytest.fixture
    def runner(self):
        return CliRunner()

    def test_project_dirs(self):
        assert project_dirs("test") == ["sql/test"]

        existing_projects = project_dirs()
        assert "sql/moz-fx-data-shared-prod" in existing_projects
        assert "sql/bigconfig.yml" not in existing_projects

    def test_metrics_render(self, tmp_path):
        file_path = tmp_path / "test_query.sql"
        file_path.write_text(
            r"""
            SELECT * FROM (
                {{ metrics.calculate(
                    metrics=['days_of_use'],
                    platform='firefox_desktop'
                ) }}
            )
        """
        )
        rendered_sql = render(file_path.name, template_folder=file_path.parent)
        assert r"{{ metrics.calculate" not in rendered_sql
        assert "days_of_use" in rendered_sql

    def test_non_existing_metrics_render(self, tmp_path):
        file_path = tmp_path / "test_query.sql"
        file_path.write_text(
            r"""
            SELECT * FROM (
                {{ metrics.calculate(
                    metrics=['not-existing'],
                    platform='firefox_desktop'
                ) }}
            )
        """
        )

        with pytest.raises(ValueError):
            render(file_path.name, template_folder=file_path.parent)

    def test_render_multiple_metrics(self, tmp_path):
        file_path = tmp_path / "test_query.sql"
        file_path.write_text(
            r"""
            SELECT * FROM (
                {{ metrics.calculate(
                    metrics=['days_of_use', 'uri_count', 'ad_clicks'],
                    platform='firefox_desktop',
                    group_by={'sample_id': 'sample_id'},
                    where='submission_date = "2023-01-01"'
                ) }}
            )
        """
        )
        rendered_sql = render(file_path.name, template_folder=file_path.parent)
        assert "metrics.calculate" not in rendered_sql
        assert r"{{" not in rendered_sql
        assert "days_of_use" in rendered_sql
        assert "clients_daily" in rendered_sql
        assert "uri_count" in rendered_sql
        assert "ad_clicks" in rendered_sql
        assert "mozdata.search.search_clients_engines_sources_daily" in rendered_sql
        assert 'submission_date = "2023-01-01"' in rendered_sql
        assert "sample_id" in rendered_sql

    def test_render_data_source(self, tmp_path):
        file_path = tmp_path / "test_query.sql"
        file_path.write_text(
            r"""
            SELECT * FROM (
                {{ metrics.data_source(
                    data_source="main",
                    platform='firefox_desktop',
                    where='submission_date = "2023-01-01"'
                ) }}
            )
        """
        )
        rendered_sql = render(file_path.name, template_folder=file_path.parent)
        assert "metrics.data_source" not in rendered_sql
        assert r"{{" not in rendered_sql
        assert "main" in rendered_sql
        assert 'submission_date = "2023-01-01"' in rendered_sql

    def test_checks_render(self, tmp_path):
        file_path = tmp_path / "checks.sql"
        file_path.write_text(
            r"""
            {{ min_row_count(1, "submission_date = @submission_date") }}
        """
        )
        kwargs = {
            "project_id": "project",
            "dataset_id": "dataset",
            "table_name": "table",
        }
        rendered_sql = render(
            file_path.name, template_folder=file_path.parent, **kwargs
        )
        assert (
            r"""{{ min_row_count(1, "submission_date = @submission_date") }}"""
            not in rendered_sql
        )
        assert "SELECT" in rendered_sql
        assert "`project.dataset.table`" in rendered_sql

    def test_qualify_table_references_in_file(self, tmp_path):
        query = "SELECT * FROM test LEFT JOIN other.joined_query"
        query_path = tmp_path / "project" / "dataset" / "test"
        query_path.mkdir(parents=True, exist_ok=True)
        query_path = query_path / "query.sql"
        query_path.write_text(query)
        result = qualify_table_references_in_file(query_path)
        expected = "SELECT * FROM `project.dataset.test` LEFT JOIN `project.other.joined_query`"
        assert result == expected

        query = "SELECT * FROM region-US.INFORMATION_SCHEMA.JOBS_BY_USER"
        query_path.write_text(query)
        result = qualify_table_references_in_file(query_path)
        expected = "SELECT * FROM `project.region-US.INFORMATION_SCHEMA.JOBS_BY_USER`"
        assert result == expected

        query = "{% if is_init() %} SELECT * FROM test {% else %} SELECT * FROM other {% endif %}"
        query_path.write_text(query)
        result = qualify_table_references_in_file(query_path)
        expected = "{% if is_init() %} SELECT * FROM `project.dataset.test` {% else %} SELECT * FROM `project.dataset.other` {% endif %}"
        assert result == expected

    def test_qualify_table_references_in_file_udf(self, tmp_path):
        udf = "SELECT * FROM test"
        udf_path = tmp_path / "project" / "dataset" / "test"
        udf_path.mkdir(parents=True, exist_ok=True)
        udf_path = udf_path / "udf.sql"
        udf_path.write_text(udf)

        with pytest.raises(NotImplementedError):
            qualify_table_references_in_file(udf_path)

    def test_qualify_table_references_in_file_scripts(self, tmp_path):
        script = "SELECT * FROM test"
        script_path = tmp_path / "project" / "dataset" / "test"
        script_path.mkdir(parents=True, exist_ok=True)
        script_path = script_path / "script.sql"
        script_path.write_text(script)

        with pytest.raises(NotImplementedError):
            qualify_table_references_in_file(script_path)

        query = "DECLARE some_var; SELECT * FROM test"
        query_path = tmp_path / "project" / "dataset" / "test"
        query_path.mkdir(parents=True, exist_ok=True)
        query_path = query_path / "query.sql"
        query_path.write_text(query)

        with pytest.raises(NotImplementedError):
            qualify_table_references_in_file(query_path)

    def test_qualify_table_references_in_file_array_fields(self, tmp_path):
        query = """
        WITH data AS (
            SELECT ARRAY_AGG(DISTINCT some_field IGNORE NULLS) AS array_field,
        )
        SELECT
            @submission_date AS submission_date,
            udf.some_udf(
                ARRAY(
                    SELECT
                        STRUCT(key, CAST(TRUE AS INT64) AS value)
                    FROM
                        data.array_field AS key
                )
            ) AS array_field
        FROM data
        """
        query_path = tmp_path / "project" / "dataset" / "test"
        query_path.mkdir(parents=True, exist_ok=True)
        query_path = query_path / "query.sql"
        query_path.write_text(query)
        result = qualify_table_references_in_file(query_path)
        expected = """
        WITH data AS (
            SELECT ARRAY_AGG(DISTINCT some_field IGNORE NULLS) AS array_field,
        )
        SELECT
            @submission_date AS submission_date,
            `project.udf.some_udf`(
                ARRAY(
                    SELECT
                        STRUCT(key, CAST(TRUE AS INT64) AS value)
                    FROM
                        data.array_field AS key
                )
            ) AS array_field
        FROM data
        """
        assert result == expected

    def test_qualify_table_references_tables(self, tmp_path):
        """Ensure tables/views are properly qualified with dataset and project id."""
        default_project = "default-project"
        default_dataset = "default_dataset"

        query_path = tmp_path / default_project / default_dataset / "test"
        query_path.mkdir(parents=True, exist_ok=True)

        input_sql = """
        WITH has_project AS (
            SELECT 1 FROM proj-ect.dataset.table1
        ),
        has_dataset AS (
            SELECT 1 FROM dataset.table1
        ),
        no_dataset AS (
            SELECT 1 AS table1 FROM table1
        ),
        has_join AS (
            SELECT
                1
            FROM
                table1
            LEFT JOIN table2
            USING (_id)
        ),
        dataset_backticks AS (
            SELECT 1 FROM `dataset.table1`
        ),
        dataset_backticks2 AS (
            SELECT 1 FROM `dataset.table1`
        ),
        alias AS (
            SELECT 1 FROM table1
        ),
        table_name_cte AS (
            SELECT 1 FROM dataset.table_name_cte
        ),
        table_name_column AS (
            SELECT table_column, table1, table2 FROM dataset.table_column
        ),
        implicit_cross_join AS (
            SELECT
                *
            FROM
                table1,
                table2
        )
        SELECT 1 FROM table_name_cte CROSS JOIN has_dataset
        """

        query_path = query_path / "query.sql"
        query_path.write_text(input_sql)

        expected = f"""
        WITH has_project AS (
            SELECT 1 FROM proj-ect.dataset.table1
        ),
        has_dataset AS (
            SELECT 1 FROM `{default_project}.dataset.table1`
        ),
        no_dataset AS (
            SELECT 1 AS table1 FROM `{default_project}.{default_dataset}.table1`
        ),
        has_join AS (
            SELECT
                1
            FROM
                `{default_project}.{default_dataset}.table1`
            LEFT JOIN `{default_project}.{default_dataset}.table2`
            USING (_id)
        ),
        dataset_backticks AS (
            SELECT 1 FROM `{default_project}.dataset.table1`
        ),
        dataset_backticks2 AS (
            SELECT 1 FROM `{default_project}.dataset.table1`
        ),
        alias AS (
            SELECT 1 FROM `{default_project}.{default_dataset}.table1`
        ),
        table_name_cte AS (
            SELECT 1 FROM `{default_project}.dataset.table_name_cte`
        ),
        table_name_column AS (
            SELECT table_column, table1, table2 FROM `{default_project}.dataset.table_column`
        ),
        implicit_cross_join AS (
            SELECT
                *
            FROM
                `{default_project}.{default_dataset}.table1`,
                `{default_project}.{default_dataset}.table2`
        )
        SELECT 1 FROM table_name_cte CROSS JOIN has_dataset
        """

        actual = qualify_table_references_in_file(query_path)

        assert actual == expected

    def test_qualify_table_references_information_schema(self, tmp_path):
        """Ensure information schema are region part is parsed correctly."""
        default_project = "default-project"
        default_dataset = "default_dataset"

        query_path = tmp_path / default_project / default_dataset / "test"
        query_path.mkdir(parents=True, exist_ok=True)

        input_sql = """
        WITH info_schema_region AS (
            SELECT 1 FROM `region-us`.INFORMATION_SCHEMA.JOBS_BY_PROJECT
        ),
        info_schema_project AS (
            SELECT 1 FROM proj.INFORMATION_SCHEMA.JOBS_BY_PROJECT
        ),
        info_schema_full AS (
            SELECT * FROM `project.region-us.INFORMATION_SCHEMA.SCHEMATA`
        ),
        info_schema_none AS (
            SELECT * FROM INFORMATION_SCHEMA.SCHEMATA
        )
        SELECT 1
        """

        query_path = query_path / "query.sql"
        query_path.write_text(input_sql)

        expected = f"""
        WITH info_schema_region AS (
            SELECT 1 FROM `{default_project}.region-us.INFORMATION_SCHEMA.JOBS_BY_PROJECT`
        ),
        info_schema_project AS (
            SELECT 1 FROM proj.INFORMATION_SCHEMA.JOBS_BY_PROJECT
        ),
        info_schema_full AS (
            SELECT * FROM `project.region-us.INFORMATION_SCHEMA.SCHEMATA`
        ),
        info_schema_none AS (
            SELECT * FROM `{default_project}.INFORMATION_SCHEMA.SCHEMATA`
        )
        SELECT 1
        """

        actual = qualify_table_references_in_file(query_path)

        assert actual == expected

    def test_qualify_table_references_udfs(self, tmp_path):
        """Ensure udf are properly qualified with dataset and project id."""
        default_project = "default-project"
        default_dataset = "default_dataset"

        query_path = tmp_path / default_project / default_dataset / "test"
        query_path.mkdir(parents=True, exist_ok=True)

        input_sql = """
        CREATE TEMP FUNCTION temp_cte() AS (
            123
        );
        SELECT
            temp_cte(),
            ARRAY_LENGTH(l),
            mozfun.dataset.func(1),
            udf.func_1(1),
            udf.nested1(udf.nested2(2)),
            `udf.func_1(3)`,
            my-project.udf.func_3(3),
            udf_js.js_func(),
        FROM
            table1
        """

        query_path = query_path / "query.sql"
        query_path.write_text(input_sql)

        expected = f"""
        CREATE TEMP FUNCTION temp_cte() AS (
            123
        );
        SELECT
            temp_cte(),
            ARRAY_LENGTH(l),
            mozfun.dataset.func(1),
            `{default_project}.udf.func_1`(1),
            `{default_project}.udf.nested1`(`{default_project}.udf.nested2`(2)),
            `{default_project}.udf.func_1`(3)`,
            my-project.udf.func_3(3),
            `{default_project}.udf_js.js_func`(),
        FROM
            `{default_project}.{default_dataset}.table1`
        """

        actual = qualify_table_references_in_file(query_path)

        assert actual == expected

    def test_extract_last_group_by_from_query_file(self, runner):
        """Test cases using a sql file path."""
        with runner.isolated_filesystem():
            test_path = (
                "sql/moz-fx-data-shared-prod/test_shredder_mitigation/test_query_v1"
            )
            os.makedirs(test_path)
            assert os.path.exists(test_path)
            assert "test_shredder_mitigation" in os.listdir(
                "sql/moz-fx-data-shared-prod"
            )
            assert is_valid_dir(None, None, test_path)

            sql_path = Path(test_path) / "query.sql"
            with open(sql_path, "w") as f:
                f.write("SELECT column_1 FROM test_table group by ALL")
            assert ["ALL"] == extract_last_group_by_from_query(sql_path=sql_path)

            with open(sql_path, "w") as f:
                f.write(
                    "SELECT column_1 FROM test_table GROUP BY (column_1) LIMIT (column_1);"
                )
            assert ["column_1"] == extract_last_group_by_from_query(sql_path=sql_path)

    def test_extract_last_group_by_from_query_sql(self):
        """Test cases using a sql text."""
        assert ["ALL"] == extract_last_group_by_from_query(
            sql_text="SELECT column_1 FROM test_table GROUP BY ALL"
        )
        assert ["1"] == extract_last_group_by_from_query(
            sql_text="SELECT column_1, SUM(metric_1) AS metric_1 FROM test_table GROUP BY 1;"
        )
        assert ["1", "2", "3"] == extract_last_group_by_from_query(
            sql_text="SELECT column_1 FROM test_table GROUP BY 1, 2, 3"
        )
        assert ["1", "2", "3"] == extract_last_group_by_from_query(
            sql_text="SELECT column_1 FROM test_table GROUP BY 1, 2, 3"
        )
        assert ["column_1", "column_2"] == extract_last_group_by_from_query(
            sql_text="""SELECT column_1, column_2 FROM test_table GROUP BY column_1, column_2 ORDER BY 1 LIMIT 100"""
        )
        assert [] == extract_last_group_by_from_query(
            sql_text="SELECT column_1 FROM test_table"
        )
        assert [] == extract_last_group_by_from_query(
            sql_text="SELECT column_1 FROM test_table;"
        )
        assert ["column_1"] == extract_last_group_by_from_query(
            sql_text="SELECT column_1 FROM test_table GROUP BY column_1"
        )
        assert ["column_1", "column_2"] == extract_last_group_by_from_query(
            sql_text="SELECT column_1, column_2 FROM test_table GROUP BY (column_1, column_2)"
        )
        assert ["column_1"] == extract_last_group_by_from_query(
            sql_text="""WITH cte AS (SELECT column_1 FROM test_table GROUP BY column_1)
            SELECT column_1 FROM cte"""
        )
        assert ["column_1"] == extract_last_group_by_from_query(
            sql_text="""WITH cte AS (SELECT column_1 FROM test_table GROUP BY column_1),
            cte2 AS (SELECT column_1, column2 FROM test_table GROUP BY column_1, column2)
            SELECT column_1 FROM cte2 GROUP BY column_1 ORDER BY 1 DESC LIMIT 1;"""
        )
        assert ["column_3"] == extract_last_group_by_from_query(
            sql_text="""WITH cte1 AS (SELECT column_1, column3 FROM test_table GROUP BY column_1, column3),
            cte3 AS (SELECT column_1, column3 FROM cte1 group by column_3) SELECT column_1 FROM cte3 limit 2;"""
        )
        assert ["column_2"] == extract_last_group_by_from_query(
            sql_text="""WITH cte1 AS (SELECT column_1 FROM test_table GROUP BY column_1),
            'cte2 AS (SELECT column_2 FROM test_table GROUP BY column_2),
            cte3 AS (SELECT column_1 FROM cte1 UNION ALL SELECT column2 FROM cte2) SELECT * FROM cte3"""
        )
        assert ["column_2"] == extract_last_group_by_from_query(
            sql_text="""WITH cte1 AS (SELECT column_1 FROM test_table GROUP BY column_1),
            cte2 AS (SELECT column_1 FROM test_table GROUP BY column_2) SELECT * FROM cte2;"""
        )

        assert ["COLUMN"] == extract_last_group_by_from_query(
            sql_text="""WITH cte1 AS (SELECT COLUMN FROM test_table GROUP BY COLUMN),
            cte2 AS (SELECT COLUMN FROM test_table GROUP BY COLUMN) SELECT * FROM cte2;"""
        )

        assert ["COLUMN"] == extract_last_group_by_from_query(
            sql_text="""WITH cte1 AS (SELECT COLUMN FROM test_table GROUP BY COLUMN),
            cte2 AS (SELECT COLUMN FROM test_table group by COLUMN) SELECT * FROM cte2;"""
        )



"""Represents a scheduled Airflow task."""

import copy
import logging
import os
import re
from enum import Enum
from fnmatch import fnmatchcase
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import attr
import cattrs
import click

from bigquery_etl.dependency import extract_table_references_without_views
from bigquery_etl.metadata.parse_metadata import Metadata, PartitionType
from bigquery_etl.query_scheduling.utils import (
    is_date_string,
    is_email,
    is_email_or_github_identity,
    is_schedule_interval,
    is_valid_dag_name,
    schedule_interval_delta,
    validate_timedelta_string,
)

AIRFLOW_TASK_TEMPLATE = "airflow_task.j2"
QUERY_FILE_RE = re.compile(
    r"^(?:.*/)?([a-zA-Z0-9_-]+)/([a-zA-Z0-9_]+)/"
    r"([a-zA-Z0-9_]+)_(v[0-9]+)/(?:query\.sql|part1\.sql|script\.sql|query\.py|checks\.sql|bigconfig\.yml)$"
)
CHECKS_FILE_RE = re.compile(
    r"^(?:.*/)?([a-zA-Z0-9_-]+)/([a-zA-Z0-9_]+)/"
    r"([a-zA-Z0-9_]+)_(v[0-9]+)/(?:checks\.sql)$"
)
DEFAULT_DESTINATION_TABLE_STR = "use-default-destination-table"
MAX_TASK_NAME_LENGTH = 250


class TriggerRule(Enum):
    """Options for task trigger rules."""

    ALL_SUCCESS = "all_success"
    ALL_FAILED = "all_failed"
    ALL_DONE = "all_done"
    ONE_FAILED = "one_failed"
    ONE_SUCCESS = "one_success"
    NONE_FAILED = "none_failed"
    NONE_SKIPPED = "none_skipped"
    DUMMY = "dummy"


class TaskParseException(Exception):
    """Raised when task scheduling config is invalid."""

    def __init__(self, message):
        """Throw TaskParseException."""
        message = f"""
        {message}

        Expected yaml format in metadata.yaml:
        scheduling:
            dag_name: string [required]
            depends_on_past: bool [optional]
            ... <more config parameters> ...
        """

        super(TaskParseException, self).__init__(message)


class UnscheduledTask(Exception):
    """Raised when a task is not scheduled."""

    pass


@attr.s(auto_attribs=True, frozen=True)
class TaskRef:
    """
    Representation of a reference to another task.

    The task can be defined in bigquery-etl or in telemetry-airflow.
    Uses attrs to simplify the class definition and provide validation.
    Docs: https://www.attrs.org
    """

    dag_name: str = attr.ib()
    task_id: str = attr.ib()
    execution_delta: Optional[str] = attr.ib(None)
    schedule_interval: Optional[str] = attr.ib(None)
    date_partition_offset: Optional[int] = attr.ib(None)
    task_group: Optional[str] = attr.ib(None)

    @property
    def task_key(self):
        """Key to uniquely identify the task."""
        return (
            f"{self.dag_name}.{self.task_group}.{self.task_id}"
            if self.task_group
            else f"{self.dag_name}.{self.task_id}"
        )

    @execution_delta.validator
    def validate_execution_delta(self, attribute, value):
        """Check that execution_delta is in a valid timedelta format."""
        if value is not None:
            validate_timedelta_string(value)

    @schedule_interval.validator
    def validate_schedule_interval(self, attribute, value):
        """Validate the schedule_interval format."""
        if value is not None and not is_schedule_interval(value):
            raise ValueError(f"Invalid schedule_interval {value}.")

    def get_execution_delta(self, schedule_interval):
        """Determine execution_delta, via schedule_interval if necessary."""
        if self.execution_delta is not None:
            return self.execution_delta
        elif self.schedule_interval is not None and schedule_interval is not None:
            execution_delta = schedule_interval_delta(
                self.schedule_interval, schedule_interval
            )
            if execution_delta != "0s":
                return execution_delta
        return None


@attr.s(auto_attribs=True, frozen=True)
class TableSensorTask:
    """Representation of a sensor task to wait for a table to exist."""

    task_id: str = attr.ib()
    table_id: str = attr.ib()
    poke_interval: Optional[str] = attr.ib(None, kw_only=True)
    timeout: Optional[str] = attr.ib(None, kw_only=True)
    retries: Optional[int] = attr.ib(None, kw_only=True)
    retry_delay: Optional[str] = attr.ib(None, kw_only=True)

    @task_id.validator
    def validate_task_id(self, attribute, value):
        """Validate the task ID."""
        if len(value) < 1 or len(value) > MAX_TASK_NAME_LENGTH:
            raise ValueError(
                f"Invalid task ID '{value}'."
                f" The task ID has to be 1 to {MAX_TASK_NAME_LENGTH} characters long."
            )
        if not re.fullmatch(r"\w+", value):
            raise ValueError(
                f"Invalid task ID '{value}'."
                f" The task ID may only contain alphanumerics and underscores."
            )

    @table_id.validator
    def validate_table_id(self, attribute, value):
        """Check that `table_id` is a fully qualified table ID."""
        if value.count(".") != 2:
            raise ValueError(
                f"Invalid table ID '{value}'."
                " Table IDs must be fully qualified with the project and dataset."
            )

    @poke_interval.validator
    def validate_poke_interval(self, attribute, value):
        """Check that `poke_interval` is a valid timedelta string."""
        if value is not None:
            validate_timedelta_string(value)

    @timeout.validator
    def validate_timeout(self, attribute, value):
        """Check that `timeout` is a valid timedelta string."""
        if value is not None:
            validate_timedelta_string(value)

    @retry_delay.validator
    def validate_retry_delay(self, attribute, value):
        """Check that `retry_delay` is a valid timedelta string."""
        if value is not None:
            validate_timedelta_string(value)


@attr.s(auto_attribs=True, frozen=True)
class TablePartitionSensorTask(TableSensorTask):
    """Representation of a sensor task to wait for a table partition to exist."""

    partition_id: str = attr.ib()


@attr.s(auto_attribs=True, frozen=True)
class FivetranTask:
    """Representation of a Fivetran data import task."""

    task_id: str = attr.ib()


class SecretDeployType(Enum):
    """Specifies how secret should be exposed in Airflow."""

    ENV = "env"
    VOLUME = "volume"


@attr.s(auto_attribs=True, frozen=True)
class Secret:
    """Represents the secret configuration used to expose credentials in the task."""

    deploy_target: str
    key: str
    deploy_type: str = attr.ib("env")
    secret: str = attr.ib("airflow-gke-secrets")

    @deploy_type.validator
    def validate_deploy_type(self, attribute, value):
        """Check that deploy_type is a valid option."""
        if value is not None and value not in set(
            deploy_type.value for deploy_type in SecretDeployType
        ):
            raise ValueError(
                f"Invalid deploy_type {value}. Needs to be either 'env' or 'volume'."
            )


@attr.s(auto_attribs=True, frozen=True)
class TaskContainerResources:
    """Represents the Kubernetes container resources configuration for a task."""

    # example dict: {"memory": "4Gi", "cpu": "500m"}
    requests: Optional[Dict[str, str]] = attr.ib(None)
    limits: Optional[Dict[str, str]] = attr.ib(None)


# Known tasks in telemetry-airflow, like stable table tasks
# https://github.com/mozilla/telemetry-airflow/blob/main/dags/copy_deduplicate.py
EXTERNAL_TASKS = {
    TaskRef(
        dag_name="copy_deduplicate",
        task_id="copy_deduplicate_main_ping",
        schedule_interval="0 1 * * *",
    ): [
        "telemetry_stable.main_v4",
        "telemetry_stable.main_v5",
        "telemetry_stable.main_use_counter_v4",
    ],
    TaskRef(
        dag_name="copy_deduplicate",
        task_id="copy_deduplicate_first_shutdown_ping",
        schedule_interval="0 1 * * *",
    ): [
        "telemetry_stable.first_shutdown_v4",
        "telemetry_stable.first_shutdown_v5",
        "telemetry_stable.first_shutdown_use_counter_v4",
    ],
    TaskRef(
        dag_name="copy_deduplicate",
        task_id="copy_deduplicate_saved_session_ping",
        schedule_interval="0 1 * * *",
    ): [
        "telemetry_stable.saved_session_v4",
        "telemetry_stable.saved_session_v5",
        "telemetry_stable.saved_session_use_counter_v4",
    ],
    TaskRef(
        dag_name="copy_deduplicate",
        task_id="bq_main_events",
        schedule_interval="0 1 * * *",
    ): ["telemetry_derived.main_events_v1"],
    TaskRef(
        dag_name="copy_deduplicate",
        task_id="event_events",
        schedule_interval="0 1 * * *",
    ): ["telemetry_derived.event_events_v1"],
    TaskRef(
        dag_name="copy_deduplicate",
        task_id="telemetry_derived__core_clients_first_seen__v1",
        schedule_interval="0 1 * * *",
    ): ["*.core_clients_first_seen*"],
    TaskRef(
        dag_name="copy_deduplicate",
        task_id="copy_deduplicate_all",
        schedule_interval="0 1 * * *",
    ): ["*_stable.*"],
}


@attr.s(auto_attribs=True)
class Task:
    """
    Representation of a task scheduled in Airflow.

    Uses attrs to simplify the class definition and provide validation.
    Docs: https://www.attrs.org
    """

    dag_name: str = attr.ib()
    query_file: str
    owner: str = attr.ib()
    email: List[str] = attr.ib([])
    task_name: Optional[str] = attr.ib(None)
    project: str = attr.ib(init=False)
    dataset: str = attr.ib(init=False)
    table: str = attr.ib(init=False)
    version: str = attr.ib(init=False)
    depends_on_past: bool = attr.ib(False)
    start_date: Optional[str] = attr.ib(None)
    date_partition_parameter: Optional[str] = "submission_date"
    table_partition_template: Optional[str] = None
    # number of days date partition parameter should be offset
    date_partition_offset: Optional[int] = None
    # indicate whether data should be published as JSON
    public_json: bool = attr.ib(False)
    # manually specified upstream dependencies
    depends_on: List[TaskRef] = attr.ib([])
    depends_on_tables_existing: List[TableSensorTask] = attr.ib([])
    depends_on_table_partitions_existing: List[TablePartitionSensorTask] = attr.ib([])
    depends_on_fivetran: List[FivetranTask] = attr.ib([])
    # task trigger rule, used to override default of "all_success"
    trigger_rule: Optional[str] = attr.ib(None)
    # manually specified downstream depdencies
    external_downstream_tasks: List[TaskRef] = attr.ib([])
    # automatically determined upstream and downstream dependencies
    upstream_dependencies: List[TaskRef] = attr.ib([])
    downstream_dependencies: List[TaskRef] = attr.ib([])
    arguments: List[str] = attr.ib([])
    parameters: List[str] = attr.ib([])
    multipart: bool = attr.ib(False)
    query_file_path: Optional[str] = None
    priority: Optional[int] = None
    referenced_tables: Optional[List[Tuple[str, str, str]]] = attr.ib(None)
    destination_table: Optional[str] = attr.ib(default=DEFAULT_DESTINATION_TABLE_STR)
    is_python_script: bool = attr.ib(False)
    is_dq_check: bool = attr.ib(False)
    # Failure of the checks task will stop the dag from executing further
    is_dq_check_fail: bool = attr.ib(True)
    is_bigeye_check: bool = attr.ib(False)
    task_concurrency: Optional[int] = attr.ib(None)
    retry_delay: Optional[str] = attr.ib(None)
    retries: Optional[int] = attr.ib(None)
    email_on_retry: Optional[bool] = attr.ib(None)
    gcp_conn_id: Optional[str] = attr.ib(None)
    gke_project_id: Optional[str] = attr.ib(None)
    gke_location: Optional[str] = attr.ib(None)
    gke_cluster_name: Optional[str] = attr.ib(None)
    query_project: Optional[str] = attr.ib(None)
    task_group: Optional[str] = attr.ib(None)
    container_resources: Optional[TaskContainerResources] = attr.ib(None)
    node_selector: Optional[Dict[str, str]] = attr.ib(None)
    startup_timeout_seconds: Optional[int] = attr.ib(None)
    secrets: Optional[List[Secret]] = attr.ib(None)
    monitoring_enabled: Optional[bool] = attr.ib(False)

    @property
    def task_key(self):
        """Key to uniquely identify the task."""
        return f"{self.dag_name}.{self.task_name}"

    @owner.validator
    def validate_owner(self, attribute, value):
        """Check that owner is a valid email address."""
        if not is_email_or_github_identity(value):
            raise ValueError(
                f"Invalid email or github identity for task owner: {value}."
            )

    @email.validator
    def validate_email(self, attribute, value):
        """Check that provided email addresses are valid."""
        if not all(map(lambda e: is_email_or_github_identity(e), value)):
            raise ValueError(f"Invalid email or github identity in DAG email: {value}.")

    @start_date.validator
    def validate_start_date(self, attribute, value):
        """Check that start_date has YYYY-MM-DD format."""
        if value is not None and not is_date_string(value):
            raise ValueError(
                f"Invalid date definition for {attribute}: {value}."
                "Dates should be specified as YYYY-MM-DD."
            )

    @dag_name.validator
    def validate_dag_name(self, attribute, value):
        """Validate the DAG name."""
        if not is_valid_dag_name(value):
            raise ValueError(
                f"Invalid DAG name {value} for task. Name must start with 'bqetl_' "
                f"or 'private_bqetl_'."
            )

    @task_name.validator
    def validate_task_name(self, attribute, value):
        """Validate the task name."""
        if value is not None:
            if len(value) < 1 or len(value) > MAX_TASK_NAME_LENGTH:
                raise ValueError(
                    f"Invalid task name {value}. "
                    f"The task name has to be 1 to {MAX_TASK_NAME_LENGTH} characters long."
                )

    @trigger_rule.validator
    def validate_trigger_rule(self, attribute, value):
        """Check that trigger_rule is a valid option."""
        if value is not None and value not in set(rule.value for rule in TriggerRule):
            raise ValueError(
                f"Invalid trigger rule {value}. "
                "See https://airflow.apache.org/docs/apache-airflow/1.10.3/concepts.html#trigger-rules for list of trigger rules"
            )

    @retry_delay.validator
    def validate_retry_delay(self, attribute, value):
        """Check that retry_delay is in a valid timedelta format."""
        if value is not None:
            validate_timedelta_string(value)

    @task_group.validator
    def validate_task_group(self, attribute, value):
        """Check that the task group name is valid."""
        if value is not None and not re.match(r"[a-zA-Z0-9_]+", value):
            raise ValueError(
                "Invalid task group identifier. Group name must match pattern [a-zA-Z0-9_]+"
            )

    def __attrs_post_init__(self):
        """Extract information from the query file name."""
        query_file_re = re.search(QUERY_FILE_RE, self.query_file)
        if query_file_re:
            self.project = query_file_re.group(1)
            self.dataset = query_file_re.group(2)
            self.table = query_file_re.group(3)
            self.version = query_file_re.group(4)

            if self.task_name is None:
                # limiting task name to allow longer dataset names
                self.task_name = f"{self.dataset}__{self.table}__{self.version}"[
                    -MAX_TASK_NAME_LENGTH:
                ]
                self.validate_task_name(None, self.task_name)

            if self.destination_table == DEFAULT_DESTINATION_TABLE_STR:
                self.destination_table = f"{self.table}_{self.version}"

            if self.destination_table is None and self.query_file_path is None:
                raise ValueError(
                    "One of destination_table or query_file_path must be specified"
                )
        else:
            raise ValueError(
                "query_file must be a path with format:"
                " sql/<project>/<dataset>/<table>_<version>"
                "/(query.sql|part1.sql|script.sql|query.py|checks.sql)"
                f" but is {self.query_file}"
            )

    @classmethod
    def of_query(cls, query_file, metadata=None, dag_collection=None):
        """
        Create task that schedules the corresponding query in Airflow.

        Raises FileNotFoundError if not metadata file exists for query.
        If `metadata` is set, then it is used instead of the metadata.yaml
        file that might exist alongside the query file.
        """
        converter = cattrs.BaseConverter()
        if metadata is None:
            metadata = Metadata.of_query_file(query_file)

        dag_name = metadata.scheduling.get("dag_name")
        if dag_name is None:
            raise UnscheduledTask(
                f"Metadata for {query_file} does not contain scheduling information."
            )

        task_config = {"query_file": str(query_file)}
        task_config.update(metadata.scheduling)

        if len(metadata.owners) <= 0:
            raise TaskParseException(
                f"No owner specified in metadata for {query_file}."
            )

        # Airflow only allows to set one owner, so we just take the first
        task_config["owner"] = metadata.owners[0]

        # Get default email from default_args if available
        default_email = []
        if dag_collection is not None:
            dag = dag_collection.dag_by_name(dag_name)
            if dag is not None:
                default_email = dag.default_args.email
        email = task_config.get("email", default_email)
        # Remove non-valid emails from owners e.g. Github identities and add to
        # Airflow email list.
        for owner in metadata.owners:
            if not is_email(owner):
                metadata.owners.remove(owner)
                click.echo(
                    f"{owner} removed from email list in DAG {metadata.scheduling['dag_name']}"
                )
        task_config["email"] = list(set(email + metadata.owners))

        # expose secret config
        task_config["secrets"] = metadata.scheduling.get("secrets", [])

        # to determine if BigEye task should be generated
        if metadata.monitoring:
            task_config["monitoring_enabled"] = metadata.monitoring.enabled

        # data processed in task should be published
        if metadata.is_public_json():
            task_config["public_json"] = True

        # Override the table_partition_template if there is no `destination_table`
        # set in the scheduling section of the metadata. If not then pass a jinja
        # template that reformats the date string used for table partition decorator.
        # See doc here for formatting conventions:
        #  https://cloud.google.com/bigquery/docs/managing-partitioned-table-data#partition_decorators
        if (
            metadata.bigquery
            and metadata.bigquery.time_partitioning
            and metadata.scheduling.get("destination_table") is None
        ):
            match metadata.bigquery.time_partitioning.type:
                case PartitionType.YEAR:
                    partition_template = '${{ dag_run.logical_date.strftime("%Y") }}'
                case PartitionType.MONTH:
                    partition_template = '${{ dag_run.logical_date.strftime("%Y%m") }}'
                case PartitionType.DAY:
                    # skip for the default case of daily partitioning
                    partition_template = None
                case PartitionType.HOUR:
                    partition_template = (
                        '${{ dag_run.logical_date.strftime("%Y%m%d%H") }}'
                    )
                case _:
                    raise TaskParseException(
                        f"Invalid partition type: {metadata.bigquery.time_partitioning.type}"
                    )

            if partition_template:
                task_config["table_partition_template"] = partition_template

        try:
            return copy.deepcopy(converter.structure(task_config, cls))
        except TypeError as e:
            raise TaskParseException(
                f"Invalid scheduling information format for {query_file}: {e}"
            )

    @classmethod
    def of_multipart_query(cls, query_file, metadata=None, dag_collection=None):
        """
        Create task that schedules the corresponding multipart query in Airflow.

        Raises FileNotFoundError if not metadata file exists for query.
        If `metadata` is set, then it is used instead of the metadata.yaml
        file that might exist alongside the query file.
        """
        task = cls.of_query(query_file, metadata, dag_collection)
        task.multipart = True
        task.query_file_path = os.path.dirname(query_file)
        return task

    @classmethod
    def of_script(cls, query_file, metadata=None, dag_collection=None):
        """
        Create task that schedules the corresponding script in Airflow.

        Raises FileNotFoundError if no metadata file exists for query.
        If `metadata` is set, then it is used instead of the metadata.yaml
        file that might exist alongside the query file.
        """
        task = cls.of_query(query_file, metadata, dag_collection)
        task.query_file_path = query_file
        task.destination_table = None
        return task

    @classmethod
    def of_python_script(cls, query_file, metadata=None, dag_collection=None):
        """
        Create a task that schedules the Python script file in Airflow.

        Raises FileNotFoundError if no metadata file exists for query.
        If `metadata` is set, then it is used instead of the metadata.yaml
        file that might exist alongside the query file.
        """
        task = cls.of_query(query_file, metadata, dag_collection)
        task.query_file_path = query_file
        task.is_python_script = True
        return task

    @classmethod
    def of_dq_check(cls, query_file, is_check_fail, metadata=None, dag_collection=None):
        """Create a task that schedules DQ check file in Airflow."""
        task = cls.of_query(query_file, metadata, dag_collection)
        task.query_file_path = query_file
        task.is_dq_check = True
        task.is_dq_check_fail = is_check_fail
        task.depends_on_past = False
        task.retries = 0
        task.depends_on_fivetran = []
        task.referenced_tables = None
        task.depends_on = []
        if task.is_dq_check_fail:
            task.task_name = (
                f"checks__fail_{task.dataset}__{task.table}__{task.version}"[
                    -MAX_TASK_NAME_LENGTH:
                ]
            )
            task.validate_task_name(None, task.task_name)
        else:
            task.task_name = (
                f"checks__warn_{task.dataset}__{task.table}__{task.version}"[
                    -MAX_TASK_NAME_LENGTH:
                ]
            )
            task.validate_task_name(None, task.task_name)
        return task

    @classmethod
    def of_bigeye_check(cls, query_file, metadata=None, dag_collection=None):
        """Create a task to trigger BigEye metric run via Airflow."""
        task = cls.of_query(query_file, metadata, dag_collection)
        task.query_file_path = None
        task.is_bigeye_check = True
        task.depends_on_past = False
        task.destination_table = None
        task.retries = 1
        task.depends_on_fivetran = []
        task.referenced_tables = None
        task.depends_on = []
        if task.is_bigeye_check:
            task.task_name = f"bigeye__{task.dataset}__{task.table}__{task.version}"[
                -MAX_TASK_NAME_LENGTH:
            ]
            task.validate_task_name(None, task.task_name)

        return task

    def to_ref(self, dag_collection):
        """Return the task as `TaskRef`."""
        return TaskRef(
            dag_name=self.dag_name,
            task_id=self.task_name,
            date_partition_offset=self.date_partition_offset,
            schedule_interval=dag_collection.dag_by_name(
                self.dag_name
            ).schedule_interval,
            task_group=self.task_group,
        )

    def _get_referenced_tables(self):
        """Use sqlglot to get tables the query depends on."""
        logging.info(f"Get dependencies for {self.task_key}")

        if self.is_python_script or self.is_bigeye_check:
            # cannot do dry runs for python scripts or BigEye config
            return self.referenced_tables or []

        if self.referenced_tables is None:
            query_files = [Path(self.query_file)]

            if self.multipart:
                # dry_run all files if query is split into multiple parts
                query_files = Path(self.query_file_path).glob("*.sql")

            table_names = {
                tuple(table.split("."))
                for query_file in query_files
                for table in extract_table_references_without_views(query_file)
            }

            # the order of table dependencies changes between requests
            # sort to maintain same order between DAG generation runs
            self.referenced_tables = sorted(table_names)
        return self.referenced_tables

    def with_upstream_dependencies(self, dag_collection):
        """Perform a dry_run to get upstream dependencies."""
        if self.upstream_dependencies:
            return

        dependencies = []

        def _duplicate_dependency(task_ref):
            return any(
                d.task_key == task_ref.task_key for d in self.depends_on + dependencies
            )

        parent_task = None
        if self.is_dq_check or self.is_bigeye_check:
            parent_task = dag_collection.task_for_table(
                self.project, self.dataset, f"{self.table}_{self.version}"
            )
            parent_task_ref = parent_task.to_ref(dag_collection)
            if not _duplicate_dependency(parent_task_ref):
                dependencies.append(parent_task_ref)

        for table in self._get_referenced_tables():
            # check if upstream task is accompanied by a check
            # the task running the check will be set as the upstream task instead
            checks_upstream_task = dag_collection.fail_checks_task_for_table(
                table[0], table[1], table[2]
            )
            bigeye_checks_upstream_task = (
                dag_collection.fail_bigeye_checks_task_for_table(
                    table[0], table[1], table[2]
                )
            )
            upstream_task = dag_collection.task_for_table(table[0], table[1], table[2])

            if upstream_task is not None:
                if upstream_task != self and upstream_task != parent_task:
                    if checks_upstream_task is not None:
                        upstream_task = checks_upstream_task

                    if bigeye_checks_upstream_task is not None:
                        upstream_task = bigeye_checks_upstream_task

                    task_ref = upstream_task.to_ref(dag_collection)
                    if not _duplicate_dependency(task_ref):
                        # Get its upstream dependencies so its date_partition_offset gets set.
                        upstream_task.with_upstream_dependencies(dag_collection)
                        task_ref = upstream_task.to_ref(dag_collection)
                        dependencies.append(task_ref)
            else:
                # see if there are some static dependencies
                for task_ref, patterns in EXTERNAL_TASKS.items():
                    if any(fnmatchcase(f"{table[1]}.{table[2]}", p) for p in patterns):
                        if not _duplicate_dependency(task_ref):
                            dependencies.append(task_ref)
                        break  # stop after the first match

        if (
            self.date_partition_parameter is not None
            and self.date_partition_offset is None
        ):
            # adjust submission_date parameter based on whether upstream tasks have
            # date partition offsets
            date_partition_offsets = [
                dependency.date_partition_offset
                for dependency in dependencies
                if dependency.date_partition_offset
            ]

            if len(date_partition_offsets) > 0:
                self.date_partition_offset = min(date_partition_offsets)
                # unset the table_partition_template property if we have an offset
                # as that will be overridden in the template via `destination_table`
                self.table_partition_template = None
                date_partition_offset_task_keys = [
                    dependency.task_key
                    for dependency in dependencies
                    if dependency.date_partition_offset == self.date_partition_offset
                ]
                logging.info(
                    f"Set {self.task_key} date partition offset"
                    f" to {self.date_partition_offset}"
                    f" based on {', '.join(date_partition_offset_task_keys)}."
                )

        self.upstream_dependencies = dependencies

    def with_downstream_dependencies(self, dag_collection):
        """Get downstream tasks by looking up upstream dependencies in DAG collection."""
        self.downstream_dependencies = [
            task_ref
            for task_ref in dag_collection.get_task_downstream_dependencies(self)
            if task_ref.dag_name != self.dag_name
        ]



"""Data review data downloaded from Bugzilla API, clean and upload to BigQuery."""

import logging
import re
from pathlib import Path

import click
import requests
import yaml
from google.cloud import bigquery

API_BASE_URL = "https://bugzilla.mozilla.org/rest/bug"
API_PARAMS = {
    "f1": "flagtypes.name",
    "o1": "substring",
    "v1": "data-review",
    "query_format": "advanced",
    "include_fields": ["id", "history"],
}

DEFAULT_PROJECT_ID = Path(__file__).parent.parent.parent.name
DEFAULT_DATASET_ID = Path(__file__).parent.parent.name
DEFAULT_TABLE_NAME = Path(__file__).parent.name

SCHEMA_FILE = Path(__file__).parent / "schema.yaml"
SCHEMA = bigquery.SchemaField.from_api_repr(
    {"name": "root", "type": "RECORD", **yaml.safe_load(SCHEMA_FILE.read_text())}
).fields


def get_api_response(url, params):
    """Get json of response if the requests to the API was successful."""
    response = requests.get(url, params)
    response.raise_for_status()
    return response.json()


def store_data_in_bigquery(data, schema, destination_project, destination_table_id):
    """Upload data to Bigquery in a single, non partitioned table."""
    client = bigquery.Client(project=destination_project)

    job_config = bigquery.LoadJobConfig(
        create_disposition="CREATE_IF_NEEDED",
        schema=schema,
        source_format=bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
        write_disposition="WRITE_TRUNCATE",
    )

    load_job = client.load_table_from_json(
        data, destination_table_id, location="US", job_config=job_config
    )
    load_job.result()
    stored_table = client.get_table(destination_table_id)
    logging.info(f"Loaded {stored_table.num_rows} rows into {destination_table_id}.")


def download_data_review_stats(url, params):
    """Download bugs with data-review from Bugzilla API and parse the data."""
    bugzilla_response = get_api_response(url, params)
    bugs = bugzilla_response["bugs"]

    data_review_rows = []
    for bug in bugs:
        bug_id = bug["id"]
        update_counter = 0
        for changeset in bug["history"]:
            update_datetime = changeset["when"][:-1]
            who = changeset["who"]
            update_counter += 1
            for change in changeset["changes"]:
                if change["field_name"] != "flagtypes.name":
                    continue
                if "data-review+" in change["added"]:
                    data_review_rows.append(
                        {
                            "steward": who,
                            "action": "approval",
                            "bug_id": bug_id,
                            "update_datetime": update_datetime,
                            "attachment_id": change.get("attachment_id"),
                            "update_counter": update_counter,
                        }
                    )
                if "data-review+" in change["removed"]:
                    data_review_rows.append(
                        {
                            "steward": who,
                            "action": "revoke approval",
                            "bug_id": bug_id,
                            "update_datetime": update_datetime,
                            "attachment_id": change.get("attachment_id"),
                            "update_counter": update_counter,
                        }
                    )
                if "data-review-" in change["added"]:
                    data_review_rows.append(
                        {
                            "steward": who,
                            "action": "rejection",
                            "bug_id": bug_id,
                            "update_datetime": update_datetime,
                            "attachment_id": change.get("attachment_id"),
                            "update_counter": update_counter,
                        }
                    )
                if "data-review-" in change["removed"]:
                    data_review_rows.append(
                        {
                            "steward": who,
                            "action": "revoke rejection",
                            "bug_id": bug_id,
                            "update_datetime": update_datetime,
                            "attachment_id": change.get("attachment_id"),
                            "update_counter": update_counter,
                        }
                    )
                if "data-review?" in change["added"]:
                    stewards = re.findall(r"data-review\?\(([^)]+)\)", change["added"])
                    for steward in stewards:
                        data_review_rows.append(
                            {
                                "steward": steward,
                                "requestor": who,
                                "action": "request",
                                "bug_id": bug_id,
                                "update_datetime": update_datetime,
                                "attachment_id": change.get("attachment_id"),
                                "update_counter": update_counter,
                            }
                        )
                if "data-review?" in change["removed"]:
                    stewards = re.findall(
                        r"data-review\?\(([^)]+)\)", change["removed"]
                    )
                    for steward in stewards:
                        data_review_rows.append(
                            {
                                "steward": steward,
                                "requestor": who,
                                "action": "remove request",
                                "bug_id": bug_id,
                                "update_datetime": update_datetime,
                                "attachment_id": change.get("attachment_id"),
                                "update_counter": update_counter,
                            }
                        )
    return data_review_rows


@click.command
@click.option(
    "--bq_project_id",
    default=DEFAULT_PROJECT_ID,
    show_default=True,
    help="BigQuery project the data is written to.",
)
@click.option(
    "--bq_dataset_id",
    default=DEFAULT_DATASET_ID,
    show_default=True,
    help="BigQuery dataset the data is written to.",
)
@click.option(
    "--bq_table_name",
    default=DEFAULT_TABLE_NAME,
    show_default=True,
    help="Bigquery table the data is written to.",
)
def run_data_review_stats(bq_project_id, bq_dataset_id, bq_table_name):
    """Download experiments from the APIs and store it in BigQuery."""
    data_review_stats = download_data_review_stats(API_BASE_URL, API_PARAMS)
    destination_table_id = f"{bq_project_id}.{bq_dataset_id}.{bq_table_name}"
    store_data_in_bigquery(
        data=data_review_stats,
        schema=SCHEMA,
        destination_project=bq_project_id,
        destination_table_id=destination_table_id,
    )


if __name__ == "__main__":
    logging.getLogger().setLevel(logging.INFO)
    run_data_review_stats()



"""Import Braze Currents from GCS into BigQuery."""

import rich_click as click
from google.cloud import bigquery


@click.command()
@click.option(
    "--destination-project",
    required=True,
    help="Google Cloud Project the table is saved to.",
)
@click.option(
    "--destination-dataset",
    required=True,
    help="Dataset in BigQuery the table is saved to.",
)
@click.option(
    "--destination-table", required=True, help="Name of the table in BigQuery."
)
@click.option("--source-bucket", required=True, help="Google Cloud Storage Bucket ")
@click.option("--source-prefix", required=True, help="Prefix of the path in GSC.")
@click.option("--event-type", required=True, help="Eventtype for the table.")
def import_braze_current_from_bucket(
    destination_project,
    destination_dataset,
    destination_table,
    source_bucket,
    source_prefix,
    event_type,
):
    """Use bigquery client to store AVRO files from bucket in BigQuery."""
    client = bigquery.Client(destination_project)
    uri = f"gs://{source_bucket}/{source_prefix}/event_type={event_type}/*"
    client.load_table_from_uri(
        uri,
        destination=f"{destination_project}.{destination_dataset}.{destination_table}",
        job_config=bigquery.LoadJobConfig(
            write_disposition=bigquery.job.WriteDisposition.WRITE_TRUNCATE,
            source_format=bigquery.job.SourceFormat.AVRO,
        ),
    ).result()



#!/usr/bin/env python3

"""
Import iProspect data from moz-fx-data-marketing-prod.iprospect.adspend_raw_v1.

The raw table represents the data in the CSV files that are updated daily and
contain the last 30 days of data. This script will import and update data for
the specified date and the 30 days prior to that date into BigQuery.
"""

from argparse import ArgumentParser
from google.cloud import bigquery
from multiprocessing.pool import ThreadPool

import datetime

parser = ArgumentParser(description=__doc__)
parser.add_argument("--date", required=True)  # expect string with format yyyy-mm-dd
parser.add_argument("--project", default="moz-fx-data-marketing-prod")
parser.add_argument("--dataset", default="iprospect")
parser.add_argument("--table", default="adspend_v1")
parser.add_argument("--parallelism", default=8)


def main():
    """Load CSV data to temporary table."""
    args = parser.parse_args()
    client = bigquery.Client(args.project)

    end_date = datetime.datetime.strptime(args.date, "%Y-%m-%d")
    date_list = [
        datetime.datetime.strftime(end_date - datetime.timedelta(days=x), "%Y-%m-%d")
        for x in range(30)
    ]

    def import_data(active_date):
        partition = active_date.replace("-", "")
        destination = f"{args.project}.{args.dataset}.{args.table}${partition}"

        job_config = bigquery.QueryJobConfig(
            destination=destination,
            time_partitioning=bigquery.TimePartitioning(
                type_=bigquery.TimePartitioningType.DAY,
                field="date",
            ),
            write_disposition="WRITE_TRUNCATE",
        )

        sql = f"""
            SELECT * EXCEPT (submission_date)
            FROM iprospect.adspend_raw_v1
            WHERE submission_date = '{args.date}'
            AND `date` = '{active_date}'
        """
        job = client.query(sql, job_config=job_config)
        print(f"Running job {job.job_id}")
        job.result()
        print(f"Updated ad spend data for {active_date}")

    with ThreadPool(args.parallelism) as pool:
        pool.map(import_data, date_list)


if __name__ == "__main__":
    main()



"""bigquery-etl CLI view command."""

import logging
import re
import string
import sys
from fnmatch import fnmatchcase
from functools import partial
from graphlib import TopologicalSorter
from multiprocessing.pool import Pool, ThreadPool
from traceback import print_exc

import rich_click as click
from google.cloud import bigquery

from ..cli.utils import (
    parallelism_option,
    paths_matching_name_pattern,
    project_id_option,
    respect_dryrun_skip_option,
    sql_dir_option,
)
from ..config import ConfigLoader
from ..dryrun import DryRun, get_credentials, get_id_token
from ..metadata.parse_metadata import METADATA_FILE, Metadata
from ..util.bigquery_id import sql_table_id
from ..util.client_queue import ClientQueue
from ..view import View, broken_views

VIEW_NAME_RE = re.compile(r"(?P<dataset>[a-zA-z0-9_]+)\.(?P<name>[a-zA-z0-9_]+)")


@click.group(help="Commands for managing views.")
def view():
    """Create the CLI group for the view command."""
    pass


@view.command(
    help="""Create a new view with name
    <dataset>.<view_name>, for example: telemetry_derived.active_profiles.
    Use the `--project_id` option to change the project the view is added to;
    default is `moz-fx-data-shared-prod`.

    Examples:

    \b
    ./bqetl view create telemetry_derived.deviations \\
      --owner=example@mozilla.com
    """,
)
@click.argument("name")
@sql_dir_option
@project_id_option(
    ConfigLoader.get("default", "project", fallback="moz-fx-data-shared-prod")
)
@click.option(
    "--owner",
    "-o",
    help="Owner of the query (email address)",
    default="example@mozilla.com",
)
def create(name, sql_dir, project_id, owner):
    """Create a new view."""
    # with dataset metadata
    try:
        match = VIEW_NAME_RE.match(name)
        name = match.group("name")
        dataset = match.group("dataset")
    except AttributeError:
        click.echo("New views must be named like: <dataset>.<view>")
        sys.exit(1)

    view = View.create(project_id, dataset, name, sql_dir)
    metadata = Metadata(
        friendly_name=string.capwords(name.replace("_", " ")),
        description="Please provide a description for the view",
        owners=[owner],
    )
    metadata.write(view.path.parent / METADATA_FILE)
    click.echo(f"Created new view {view.path}")


@view.command(
    help="""Validate a view.
    Checks formatting, naming, references and dry runs the view.

    Examples:

    ./bqetl view validate telemetry.clients_daily
    """
)
@click.argument("name", required=False)
@sql_dir_option
@project_id_option(default=None)
@parallelism_option()
def validate(
    name,
    sql_dir,
    project_id,
    parallelism,
):
    """Validate the view definition."""
    view_files = paths_matching_name_pattern(
        name, sql_dir, project_id, files=("view.sql",)
    )
    id_token = get_id_token()
    views = [View.from_file(f, id_token=id_token) for f in view_files]

    with Pool(parallelism) as p:
        result = p.map(_view_is_valid, views)
    if not all(result):
        sys.exit(1)

    click.echo("All views are valid.")


def _view_is_valid(v: View) -> bool:
    return v.is_valid()


@view.command(
    help="""Publish views.
    Examples:

    # Publish all views
    ./bqetl view publish

    # Publish a specific view
    ./bqetl view publish telemetry.clients_daily
    """
)
@click.argument("name", required=False)
@sql_dir_option
@project_id_option(default=None)
@click.option(
    "--target-project",
    help=(
        "If specified, create views in the target project rather than"
        " the project specified in the file. Only views for "
        " moz-fx-data-shared-prod will be published if this is set."
    ),
)
@click.option("--log-level", default="INFO", help="Defaults to INFO")
@parallelism_option()
@click.option(
    "--dry_run",
    "--dry-run",
    is_flag=True,
    help="Validate view definitions, but do not publish them.",
)
@click.option(
    "--user-facing-only",
    "--user_facing_only",
    is_flag=True,
    help=(
        "Publish user-facing views only. User-facing views are views"
        " part of datasets without suffixes (such as telemetry,"
        " but not telemetry_derived)."
    ),
)
@click.option(
    "--skip-authorized",
    "--skip_authorized",
    is_flag=True,
    help="Don't publish views with labels: {authorized: true} in metadata.yaml",
)
@click.option(
    "--force",
    is_flag=True,
    help="Publish views even if there are no changes to the view query",
)
@click.option(
    "--add-managed-label",
    "--add_managed_label",
    is_flag=True,
    help=(
        'Add a label "managed" to views, that can be used to remove views from BigQuery'
        " when they are removed from --sql-dir."
    ),
)
@respect_dryrun_skip_option(default=False)
def publish(
    name,
    sql_dir,
    project_id,
    target_project,
    log_level,
    parallelism,
    dry_run,
    user_facing_only,
    skip_authorized,
    force,
    add_managed_label,
    respect_dryrun_skip,
):
    """Publish views."""
    # set log level
    try:
        logging.basicConfig(level=log_level, format="%(levelname)s %(message)s")
    except ValueError as e:
        raise click.ClickException(f"argument --log-level: {e}")
    credentials = get_credentials()

    views = _collect_views(name, sql_dir, project_id, user_facing_only, skip_authorized)
    if respect_dryrun_skip:
        views = [view for view in views if view.path not in DryRun.skipped_files()]
    if add_managed_label:
        for view in views:
            view.labels["managed"] = ""
    if not force:
        has_changes = partial(_view_has_changes, target_project, credentials)

        # only views with changes
        with Pool(parallelism) as p:
            changes = p.map(has_changes, views)
        views = [v for v, has_changes in zip(views, changes) if has_changes]
    views_by_id = {v.view_identifier: v for v in views}

    view_id_graph = {
        view.view_identifier: {
            ref for ref in view.table_references if ref in views_by_id
        }
        for view in views
    }

    view_id_order = TopologicalSorter(view_id_graph).static_order()

    client = bigquery.Client(credentials=credentials)

    result = []
    for view_id in view_id_order:
        try:
            result.append(views_by_id[view_id].publish(target_project, dry_run, client))
        except Exception:
            print(f"Failed to publish view: {view_id}")
            print_exc()
            result.append(False)

    if not all(result):
        sys.exit(1)

    click.echo("All have been published.")


def _view_has_changes(target_project, credentials, view):
    return view.has_changes(target_project, credentials)


def _collect_views(name, sql_dir, project_id, user_facing_only, skip_authorized):
    view_files = paths_matching_name_pattern(
        name, sql_dir, project_id, files=("view.sql",)
    )
    id_token = get_id_token()

    views = [View.from_file(f, id_token=id_token) for f in view_files]
    if user_facing_only:
        views = [v for v in views if v.is_user_facing]
    if skip_authorized:
        views = [
            v
            for v in views
            if not (
                v.metadata
                and v.metadata.labels
                # labels with boolean true are translated to ""
                and v.metadata.labels.get("authorized") == ""
            )
        ]
    return views


@view.command(
    help="""Remove managed views that are not present in the sql dir.
    Examples:

    # Clean managed views in shared prod
    ./bqetl view clean --target-project=moz-fx-data-shared-prod --skip-authorized

    # Clean managed user facing views in mozdata
    ./bqetl view clean --target-project=mozdata --user-facing-only --skip-authorized
    """
)
@click.argument("name", required=False)
@sql_dir_option
@project_id_option(default=None)
@click.option(
    "--target-project",
    help=(
        "If specified, clean views in the target project rather than"
        " the project specified in the file. Only views for "
        " moz-fx-data-shared-prod will be included if this is set."
    ),
)
@click.option("--log-level", default="INFO", help="Defaults to INFO")
@parallelism_option()
@click.option(
    "--dry_run",
    "--dry-run",
    is_flag=True,
    help="Identify views to delete, but do not delete them.",
)
@click.option(
    "--user-facing-only",
    "--user_facing_only",
    is_flag=True,
    help=(
        "Remove user-facing views only. User-facing views are views"
        " part of datasets without suffixes (such as telemetry,"
        " but not telemetry_derived)."
    ),
)
@click.option(
    "--skip-authorized",
    "--skip_authorized",
    is_flag=True,
    help="Don't publish views with labels: {authorized: true} in metadata.yaml",
)
def clean(
    name,
    sql_dir,
    project_id,
    target_project,
    log_level,
    parallelism,
    dry_run,
    user_facing_only,
    skip_authorized,
):
    """Clean managed views."""
    # set log level
    try:
        logging.basicConfig(level=log_level, format="%(levelname)s %(message)s")
    except ValueError as e:
        raise click.ClickException(f"argument --log-level: {e}")

    if project_id is None and target_project is None:
        raise click.ClickException("command requires --project-id or --target-project")

    expected_view_ids = {
        view.target_view_identifier(target_project)
        for view in _collect_views(
            name, sql_dir, project_id, user_facing_only, skip_authorized
        )
    }

    client_q = ClientQueue([project_id], parallelism)
    with client_q.client() as client:
        datasets = [
            dataset
            for dataset in client.list_datasets(target_project)
            if not user_facing_only
            or not dataset.dataset_id.endswith(
                tuple(
                    ConfigLoader.get(
                        "default", "non_user_facing_dataset_suffixes", fallback=[]
                    )
                )
            )
        ]

    with ThreadPool(parallelism) as p:
        managed_view_ids = {
            view
            for views in p.starmap(
                client_q.with_client,
                (
                    (_list_managed_views, dataset, name, skip_authorized)
                    for dataset in datasets
                ),
                chunksize=1,
            )
            for view in views
        }

        remove_view_ids = sorted(managed_view_ids - expected_view_ids)
        p.starmap(
            client_q.with_client,
            ((_remove_view, view_id, dry_run) for view_id in remove_view_ids),
            chunksize=1,
        )


def _list_managed_views(client, dataset, pattern, skip_authorized):
    query = f"""
      SELECT
        table_catalog || "." || table_schema || "." || table_name AS table_id,
        CONTAINS_SUBSTR(option_value, 'STRUCT("authorized", "")') AS is_authorized
      FROM
        `{dataset.project}.{dataset.dataset_id}.INFORMATION_SCHEMA.VIEWS`
      INNER JOIN
        `{dataset.project}.{dataset.dataset_id}.INFORMATION_SCHEMA.TABLE_OPTIONS`
      USING
        (table_catalog,
         table_schema,
         table_name)
      WHERE
        option_name = "labels"
      AND CONTAINS_SUBSTR(option_value, 'STRUCT("managed", "")')
    """

    # running a query against information schema instead of using the API to list tables is much faster
    job = client.query(query)
    result = list(job.result())
    return [
        row.table_id
        for row in result
        if (pattern is None or fnmatchcase(sql_table_id(row.table_id), f"*{pattern}"))
        and (not skip_authorized or not row.is_authorized)
    ]


def _remove_view(client, view_id, dry_run):
    if dry_run:
        click.echo(f"Would delete {view_id}")
    else:
        click.echo(f"Deleting {view_id}")
        client.delete_table(view_id)


@view.command(
    help="""List broken views.
    Examples:

    # Publish all views
    ./bqetl view list-broken

    # Publish a specific view
    ./bqetl view list-broken --only telemetry
    """
)
@project_id_option()
@parallelism_option()
@click.option(
    "--only",
    "-o",
    help="Process only the given tables",
)
@click.option(
    "--log-level",
    "--log_level",
    help="Log level.",
    default=logging.getLevelName(logging.INFO),
    type=str.upper,
)
def list_broken(project_id, parallelism, only, log_level):
    """List broken views."""
    broken_views.list_broken_views(project_id, parallelism, only, log_level)



"""Delete user data from long term storage."""

import logging
import warnings
from argparse import ArgumentParser
from collections import defaultdict
from dataclasses import dataclass, replace
from datetime import datetime, timedelta
from functools import partial
from multiprocessing.pool import ThreadPool
from operator import attrgetter
from textwrap import dedent
from typing import Callable, Iterable, Optional, Tuple, Union

from google.api_core.exceptions import NotFound
from google.cloud import bigquery
from google.cloud.bigquery import QueryJob

from ..format_sql.formatter import reformat
from ..util import standard_args
from ..util.bigquery_id import FULL_JOB_ID_RE, full_job_id, sql_table_id
from ..util.client_queue import ClientQueue
from ..util.exceptions import BigQueryInsertError
from .config import (
    DELETE_TARGETS,
    DeleteSource,
    DeleteTarget,
    find_experiment_analysis_targets,
    find_glean_targets,
    find_pioneer_targets,
)

NULL_PARTITION_ID = "__NULL__"
OUTSIDE_RANGE_PARTITION_ID = "__UNPARTITIONED__"

parser = ArgumentParser(description=__doc__)
standard_args.add_dry_run(parser)
parser.add_argument(
    "--environment",
    default="telemetry",
    const="telemetry",
    nargs="?",
    choices=["telemetry", "pioneer", "experiments"],
    help="environment to run in (dictates the choice of source and target tables): "
    "telemetry - standard environment, "
    "pioneer - restricted pioneer environment, "
    "experiments - experiment analysis tables",
)
parser.add_argument(
    "--pioneer-study-projects",
    "--pioneer_study_projects",
    default=[],
    help="Pioneer study-specific analysis projects to include in data deletion.",
    type=lambda s: [i for i in s.split(",")],
)
parser.add_argument(
    "--partition-limit",
    "--partition_limit",
    metavar="N",
    type=int,
    help="Only use the first N partitions per table; requires --dry-run",
)
parser.add_argument(
    "--no-use-dml",
    "--no_use_dml",
    action="store_false",
    dest="use_dml",
    help="Use SELECT * FROM instead of DELETE in queries to avoid concurrent DML limit "
    "or errors due to read-only permissions being insufficient to dry run DML; unless "
    "used with --dry-run, DML will still be used for special partitions like __NULL__",
)
standard_args.add_log_level(parser)
standard_args.add_parallelism(parser)
parser.add_argument(
    "-e",
    "--end-date",
    "--end_date",
    default=datetime.utcnow().date(),
    type=lambda x: datetime.strptime(x, "%Y-%m-%d").date(),
    help="last date of pings to delete; One day after last date of "
    "deletion requests to process; defaults to today in UTC",
)
parser.add_argument(
    "-s",
    "--start-date",
    "--start_date",
    type=lambda x: datetime.strptime(x, "%Y-%m-%d").date(),
    help="first date of deletion requests to process; DOES NOT apply to ping date; "
    "defaults to 14 days before --end-date in UTC",
)
standard_args.add_billing_projects(parser, default=["moz-fx-data-bq-batch-prod"])
parser.add_argument(
    "--source-project",
    "--source_project",
    help="override the project used for deletion request tables",
)
parser.add_argument(
    "--target-project",
    "--target_project",
    help="override the project used for target tables",
)
parser.add_argument(
    "--max-single-dml-bytes",
    "--max_single_dml_bytes",
    default=10 * 2**40,
    type=int,
    help="Maximum number of bytes in a table that should be processed using a single "
    "DML query; tables above this limit will be processed using per-partition "
    "queries; this option prevents queries against large tables from exceeding the "
    "6-hour time limit; defaults to 10 TiB",
)
standard_args.add_priority(parser)
parser.add_argument(
    "--state-table",
    "--state_table",
    metavar="TABLE",
    help="Table for recording state; Used to avoid repeating deletes if interrupted; "
    "Create it if it does not exist; By default state is not recorded",
)
parser.add_argument(
    "--task-table",
    "--task_table",
    metavar="TABLE",
    help="Table for recording tasks; Used along with --state-table to determine "
    "progress; Create it if it does not exist; By default tasks are not recorded",
)
standard_args.add_table_filter(parser)
parser.add_argument(
    "--sampling-tables",
    "--sampling_tables",
    nargs="+",
    dest="sampling_tables",
    help="Create tasks per sample id for the given table(s).  Table format is dataset.table_name.",
    default=[],
)
parser.add_argument(
    "--sampling-parallelism",
    "--sampling_parallelism",
    type=int,
    default=10,
    help="Number of concurrent queries to run per partition when shredding per sample id",
)
parser.add_argument(
    "--temp-dataset",
    "--temp_dataset",
    metavar="PROJECT.DATASET",
    help="Dataset (project.dataset format) to write intermediate results of sampled queries to. "
    "Must be specified when --sampling-tables is set.",
)


@dataclass
class DeleteJobResults:
    """Subset of a bigquery job object retaining only fields that are needed."""

    job_id: str
    total_bytes_processed: Optional[int]
    num_dml_affected_rows: Optional[int]
    destination: str


def record_state(client, task_id, job, dry_run, start_date, end_date, state_table):
    """Record the job for task_id in state_table."""
    if state_table is not None:
        job_id = "a job_id" if dry_run else full_job_id(job)
        insert_tense = "Would insert" if dry_run else "Inserting"
        logging.info(f"{insert_tense} {job_id} in {state_table} for task: {task_id}")
        if not dry_run:
            BigQueryInsertError.raise_if_present(
                errors=client.insert_rows_json(
                    state_table,
                    [
                        {
                            "task_id": task_id,
                            "job_id": job_id,
                            "job_created": job.created.isoformat(),
                            "start_date": start_date.isoformat(),
                            "end_date": end_date.isoformat(),
                        }
                    ],
                    skip_invalid_rows=False,
                )
            )


def wait_for_job(
    client,
    states,
    task_id,
    dry_run,
    create_job,
    check_table_existence=False,
    **state_kwargs,
) -> DeleteJobResults:
    """Get a job from state or create a new job, and wait for the job to complete."""
    job = None
    if task_id in states:
        job = client.get_job(**FULL_JOB_ID_RE.fullmatch(states[task_id]).groupdict())  # type: ignore[union-attr]
        if job.errors:
            logging.info(f"Previous attempt failed, retrying for {task_id}")
            job = None
        elif job.ended:
            # if destination table no longer exists (temp table expired), rerun job
            try:
                if check_table_existence:
                    client.get_table(job.destination)
                logging.info(
                    f"Previous attempt succeeded, reusing result for {task_id}"
                )
            except NotFound:
                logging.info(f"Previous result expired, retrying for {task_id}")
                job = None
        else:
            logging.info(f"Previous attempt still running for {task_id}")
    if job is None:
        job = create_job(client)
        record_state(
            client=client, task_id=task_id, dry_run=dry_run, job=job, **state_kwargs
        )
    if not dry_run and not job.ended:
        logging.info(f"Waiting on {full_job_id(job)} for {task_id}")
        job.result()

    try:
        bytes_processed = job.total_bytes_processed
    except AttributeError:
        bytes_processed = 0

    return DeleteJobResults(
        job_id=job.job_id,
        total_bytes_processed=bytes_processed,
        num_dml_affected_rows=(
            job.num_dml_affected_rows if isinstance(job, QueryJob) else None
        ),
        destination=job.destination,
    )


def get_task_id(target, partition_id):
    """Get unique task id for state tracking."""
    task_id = sql_table_id(target)
    if partition_id:
        task_id += f"${partition_id}"
    return task_id


@dataclass
class Partition:
    """Return type for get_partition."""

    condition: str
    id: Optional[str] = None
    is_special: bool = False


def delete_from_partition(
    dry_run: bool,
    partition: Partition,
    priority: str,
    source_condition: str,
    sources: Iterable[DeleteSource],
    target: DeleteTarget,
    use_dml: bool,
    sample_id: Optional[int] = None,
    temp_dataset: Optional[str] = None,
    clustering_fields: Optional[Iterable[str]] = None,
    **wait_for_job_kwargs,
):
    """Return callable to handle deletion requests for partitions of a target table."""
    job_config = bigquery.QueryJobConfig(dry_run=dry_run, priority=priority)
    # whole table operations must use DML to protect against dropping partitions in the
    # case of conflicting write operations in ETL, and special partitions must use DML
    # because they can't be set as a query destination.
    if partition.id is None or partition.is_special:
        use_dml = True
    elif sample_id is not None:
        use_dml = False
        job_config.destination = (
            f"{temp_dataset}.{target.table_id}_{partition.id}__sample_{sample_id}"
        )
        job_config.write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE
        job_config.clustering_fields = clustering_fields
    elif not use_dml:
        job_config.destination = f"{sql_table_id(target)}${partition.id}"
        job_config.write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE

    def create_job(client) -> bigquery.QueryJob:
        if use_dml:
            field_condition = " OR ".join(
                f"""
                {field} IN (
                  SELECT
                    {source.field}
                  FROM
                    `{sql_table_id(source)}`
                  WHERE
                """
                + " AND ".join((source_condition, *source.conditions))
                + ")"
                for field, source in zip(target.fields, sources)
            )
            query = reformat(
                f"""
                DELETE
                  `{sql_table_id(target)}`
                WHERE
                  ({field_condition})
                  AND ({partition.condition})
                """
            )
        else:
            field_joins = "".join(
                (
                    f"""
                LEFT JOIN
                  (
                    SELECT
                      {source.field} AS _source_{index}
                    FROM
                      `{sql_table_id(source)}`
                    WHERE
                """
                    + " AND ".join((source_condition, *source.conditions))
                    + (f" AND sample_id = {sample_id}" if sample_id is not None else "")
                    + f"""
                  )
                  ON {field} = _source_{index}
                """
                )
                for index, (field, source) in enumerate(zip(target.fields, sources))
            )
            field_conditions = " AND ".join(
                f"_source_{index} IS NULL" for index, _ in enumerate(sources)
            )

            if partition.id is None:
                # only apply field conditions on partition.condition
                field_conditions = f"""
                ({partition.condition}) IS NOT TRUE
                OR ({field_conditions})
                """
                # always true partition condition to satisfy require_partition_filter
                partition_condition = f"""
                ({partition.condition}) IS NOT TRUE
                OR ({partition.condition})
                """
            else:
                partition_condition = partition.condition

            query = reformat(
                f"""
                SELECT
                  _target.*,
                FROM
                  `{sql_table_id(target)}` AS _target
                {field_joins}
                WHERE
                  ({field_conditions})
                  AND ({partition_condition})
                  {f" AND sample_id = {sample_id}" if sample_id is not None else ""}
                """
            )
        run_tense = "Would run" if dry_run else "Running"
        logging.debug(f"{run_tense} query: {query}")
        return client.query(query, job_config=job_config)

    return partial(
        wait_for_job, create_job=create_job, dry_run=dry_run, **wait_for_job_kwargs
    )


def delete_from_partition_with_sampling(
    dry_run: bool,
    partition: Partition,
    priority: str,
    source_condition: str,
    sources: Iterable[DeleteSource],
    target: DeleteTarget,
    use_dml: bool,
    sampling_parallelism: int,
    temp_dataset: str,
    **wait_for_job_kwargs,
):
    """Return callable to delete from a partition of a target table per sample id."""
    copy_job_config = bigquery.CopyJobConfig(
        write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
        create_disposition=bigquery.CreateDisposition.CREATE_IF_NEEDED,
    )
    target_table = f"{sql_table_id(target)}${partition.id}"

    def delete_by_sample(client) -> Union[bigquery.CopyJob, bigquery.QueryJob]:
        intermediate_clustering_fields = client.get_table(
            target_table
        ).clustering_fields

        tasks = [
            delete_from_partition(
                dry_run=dry_run,
                partition=partition,
                priority=priority,
                source_condition=source_condition,
                sources=sources,
                target=target,
                use_dml=use_dml,
                temp_dataset=temp_dataset,
                sample_id=s,
                clustering_fields=intermediate_clustering_fields,
                check_table_existence=True,
                **{
                    **wait_for_job_kwargs,
                    # override task id with sample id suffix
                    "task_id": f"{wait_for_job_kwargs['task_id']}__sample_{s}",
                },
            )
            for s in range(100)
        ]

        # Run all 100 delete functions in parallel, exception is raised without retry if any fail
        with ThreadPool(sampling_parallelism) as pool:
            jobs = [
                pool.apply_async(
                    task,
                    args=(client,),
                )
                for task in tasks
            ]

            results = [job.get() for job in jobs]

        intermediate_tables = [result.destination for result in results]
        run_tense = "Would copy" if dry_run else "Copying"
        logging.debug(
            f"{run_tense} {len(intermediate_tables)} "
            f"{[str(t) for t in intermediate_tables]} to {target_table}"
        )
        if not dry_run:
            copy_job = client.copy_table(
                sources=intermediate_tables,
                destination=target_table,
                job_config=copy_job_config,
            )
            # simulate query job properties for logging
            copy_job.total_bytes_processed = sum(
                [r.total_bytes_processed for r in results]
            )
            return copy_job
        else:
            # copy job doesn't have dry runs so dry run base partition for byte estimate
            return client.query(
                f"SELECT * FROM `{sql_table_id(target)}` WHERE {partition.condition}",
                job_config=bigquery.QueryJobConfig(dry_run=True),
            )

    return partial(
        wait_for_job,
        create_job=delete_by_sample,
        dry_run=dry_run,
        **wait_for_job_kwargs,
    )


def get_partition_expr(table):
    """Get the SQL expression to use for a table's partitioning field."""
    if table.range_partitioning:
        return table.range_partitioning.field
    if table.time_partitioning:
        return f"CAST({table.time_partitioning.field or '_PARTITIONTIME'} AS DATE)"


def get_partition(table, partition_expr, end_date, id_=None) -> Optional[Partition]:
    """Return a Partition for id_ unless it is a date on or after end_date."""
    if id_ is None:
        if table.time_partitioning:
            return Partition(condition=f"{partition_expr} < '{end_date}'")
        return Partition(condition="TRUE")
    if id_ == NULL_PARTITION_ID:
        if table.time_partitioning:
            return Partition(
                condition=f"{table.time_partitioning.field} IS NULL",
                id=id_,
                is_special=True,
            )
        return Partition(condition=f"{partition_expr} IS NULL", id=id_, is_special=True)
    if table.time_partitioning:
        date = datetime.strptime(id_, "%Y%m%d").date()
        if date < end_date:
            return Partition(f"{partition_expr} = '{date}'", id_)
        return None
    if table.range_partitioning:
        if id_ == OUTSIDE_RANGE_PARTITION_ID:
            return Partition(
                condition=f"{partition_expr} < {table.range_partitioning.range_.start} "
                f"OR {partition_expr} >= {table.range_partitioning.range_.end}",
                id=id_,
                is_special=True,
            )
        if table.range_partitioning.range_.interval > 1:
            return Partition(
                condition=f"{partition_expr} BETWEEN {id_} "
                f"AND {int(id_) + table.range_partitioning.range_.interval - 1}",
                id=id_,
            )
    return Partition(condition=f"{partition_expr} = {id_}", id=id_)


def list_partitions(
    client, table, partition_expr, end_date, max_single_dml_bytes, partition_limit
):
    """List the relevant partitions in a table."""
    partitions = [
        partition
        for partition in (
            [
                get_partition(table, partition_expr, end_date, row["partition_id"])
                for row in client.query(
                    dedent(
                        f"""
                        SELECT
                          partition_id
                        FROM
                          [{sql_table_id(table)}$__PARTITIONS_SUMMARY__]
                        """
                    ).strip(),
                    bigquery.QueryJobConfig(use_legacy_sql=True),
                ).result()
            ]
            if table.num_bytes > max_single_dml_bytes and partition_expr is not None
            else [get_partition(table, partition_expr, end_date)]
        )
        if partition is not None
    ]
    if partition_limit:
        return sorted(partitions, key=attrgetter("id"), reverse=True)[:partition_limit]
    return partitions


@dataclass
class Task:
    """Return type for delete_from_table."""

    table: bigquery.Table
    sources: Tuple[DeleteSource]
    partition_id: Optional[str]
    func: Callable[[bigquery.Client], bigquery.QueryJob]

    @property
    def partition_sort_key(self):
        """Return a tuple to control the order in which tasks will be handled.

        When used with reverse=True, handle tasks without partition_id, then
        tasks without time_partitioning, then most recent dates first.
        """
        return (
            self.partition_id is None,
            self.table.time_partitioning is None,
            self.partition_id,
        )


def delete_from_table(
    client,
    target,
    sources,
    dry_run,
    end_date,
    max_single_dml_bytes,
    partition_limit,
    sampling_parallelism,
    use_sampling,
    temp_dataset,
    **kwargs,
) -> Iterable[Task]:
    """Yield tasks to handle deletion requests for a target table."""
    try:
        table = client.get_table(sql_table_id(target))
    except NotFound:
        logging.warning(f"Skipping {sql_table_id(target)} due to NotFound exception")
        return ()
    partition_expr = get_partition_expr(table)
    for partition in list_partitions(
        client, table, partition_expr, end_date, max_single_dml_bytes, partition_limit
    ):
        # no sampling for __NULL__ partition
        if use_sampling and not partition.is_special:
            kwargs["sampling_parallelism"] = sampling_parallelism
            delete_func: Callable = delete_from_partition_with_sampling
        else:
            if use_sampling:
                logging.warning(
                    "Cannot use sampling on full table deletion, "
                    f"{target.dataset_id}.{target.table_id} is too small to use sampling"
                )
            kwargs.pop("sampling_parallelism", None)
            delete_func = delete_from_partition

        yield Task(
            table=table,
            sources=sources,
            partition_id=partition.id,
            func=delete_func(
                dry_run=dry_run,
                partition=partition,
                target=target,
                sources=sources,
                task_id=get_task_id(target, partition.id),
                end_date=end_date,
                temp_dataset=temp_dataset,
                **kwargs,
            ),
        )


def main():
    """Process deletion requests."""
    args = parser.parse_args()
    if args.partition_limit is not None and not args.dry_run:
        parser.print_help()
        logging.warning("ERROR: --partition-limit specified without --dry-run")
    if len(args.sampling_tables) > 0 and args.temp_dataset is None:
        parser.error("--temp-dataset must be specified when using --sampling-tables")
    if args.start_date is None:
        args.start_date = args.end_date - timedelta(days=14)
    source_condition = (
        f"DATE(submission_timestamp) >= '{args.start_date}' "
        f"AND DATE(submission_timestamp) < '{args.end_date}'"
    )
    client_q = ClientQueue(
        args.billing_projects,
        args.parallelism,
        connection_pool_max_size=(
            max(args.parallelism * args.sampling_parallelism, 12)
            if len(args.sampling_tables) > 0
            else None
        ),
    )
    client = client_q.default_client
    states = {}
    if args.state_table:
        state_table_exists = False
        try:
            client.get_table(args.state_table)
            state_table_exists = True
        except NotFound:
            if not args.dry_run:
                client.create_table(
                    bigquery.Table(
                        args.state_table,
                        [
                            bigquery.SchemaField("task_id", "STRING"),
                            bigquery.SchemaField("job_id", "STRING"),
                            bigquery.SchemaField("job_created", "TIMESTAMP"),
                            bigquery.SchemaField("start_date", "DATE"),
                            bigquery.SchemaField("end_date", "DATE"),
                        ],
                    )
                )
                state_table_exists = True
        if state_table_exists:
            states = dict(
                client.query(
                    reformat(
                        f"""
                        SELECT
                          task_id,
                          job_id,
                        FROM
                          `{args.state_table}`
                        WHERE
                          end_date = '{args.end_date}'
                        ORDER BY
                          job_created
                        """
                    )
                ).result()
            )

    if args.environment == "telemetry":
        with ThreadPool(6) as pool:
            glean_targets = find_glean_targets(pool, client)
        targets_with_sources = (
            *DELETE_TARGETS.items(),
            *glean_targets.items(),
        )
    elif args.environment == "experiments":
        targets_with_sources = find_experiment_analysis_targets(client).items()
    elif args.environment == "pioneer":
        with ThreadPool(args.parallelism) as pool:
            targets_with_sources = find_pioneer_targets(
                pool, client, study_projects=args.pioneer_study_projects
            ).items()

    missing_sampling_tables = [
        t
        for t in args.sampling_tables
        if t not in [target.table for target, _ in targets_with_sources]
    ]
    if len(missing_sampling_tables) > 0:
        raise ValueError(
            f"{len(missing_sampling_tables)} sampling tables not found in "
            f"targets: {missing_sampling_tables}"
        )

    tasks = [
        task
        for target, sources in targets_with_sources
        if args.table_filter(target.table)
        for task in delete_from_table(
            client=client,
            target=replace(target, project=args.target_project or target.project),
            sources=[
                replace(source, project=args.source_project or source.project)
                for source in (sources if isinstance(sources, tuple) else (sources,))
            ],
            source_condition=source_condition,
            dry_run=args.dry_run,
            use_dml=args.use_dml,
            priority=args.priority,
            start_date=args.start_date,
            end_date=args.end_date,
            max_single_dml_bytes=args.max_single_dml_bytes,
            partition_limit=args.partition_limit,
            state_table=args.state_table,
            states=states,
            sampling_parallelism=args.sampling_parallelism,
            use_sampling=target.table in args.sampling_tables,
            temp_dataset=args.temp_dataset,
        )
    ]

    if not tasks:
        logging.error("No tables selected")
        parser.exit(1)
    # ORDER BY partition_sort_key DESC, sql_table_id ASC
    # https://docs.python.org/3/howto/sorting.html#sort-stability-and-complex-sorts
    tasks.sort(key=lambda task: sql_table_id(task.table))
    tasks.sort(key=attrgetter("partition_sort_key"), reverse=True)

    with ThreadPool(args.parallelism) as pool:
        if args.task_table and not args.dry_run:
            # record task information
            try:
                client.get_table(args.task_table)
            except NotFound:
                table = bigquery.Table(
                    args.task_table,
                    [
                        bigquery.SchemaField("task_id", "STRING"),
                        bigquery.SchemaField("start_date", "DATE"),
                        bigquery.SchemaField("end_date", "DATE"),
                        bigquery.SchemaField("target", "STRING"),
                        bigquery.SchemaField("target_rows", "INT64"),
                        bigquery.SchemaField("target_bytes", "INT64"),
                        bigquery.SchemaField("source_bytes", "INT64"),
                    ],
                )
                table.time_partitioning = bigquery.TimePartitioning()
                client.create_table(table)
            sources = list(set(source for task in tasks for source in task.sources))
            source_bytes = {
                source: job.total_bytes_processed
                for source, job in zip(
                    sources,
                    pool.starmap(
                        client.query,
                        [
                            (
                                reformat(
                                    f"""
                                    SELECT
                                      {source.field}
                                    FROM
                                      `{sql_table_id(source)}`
                                    WHERE
                                      {source_condition}
                                    """
                                ),
                                bigquery.QueryJobConfig(dry_run=True),
                            )
                            for source in sources
                        ],
                        chunksize=1,
                    ),
                )
            }
            step = 10000  # max 10K rows per insert
            for start in range(0, len(tasks), step):
                end = start + step
                BigQueryInsertError.raise_if_present(
                    errors=client.insert_rows_json(
                        args.task_table,
                        [
                            {
                                "task_id": get_task_id(task.table, task.partition_id),
                                "start_date": args.start_date.isoformat(),
                                "end_date": args.end_date.isoformat(),
                                "target": sql_table_id(task.table),
                                "target_rows": task.table.num_rows,
                                "target_bytes": task.table.num_bytes,
                                "source_bytes": sum(
                                    map(source_bytes.get, task.sources)
                                ),
                            }
                            for task in tasks[start:end]
                        ],
                    )
                )
        results = pool.map(
            client_q.with_client, (task.func for task in tasks), chunksize=1
        )
    jobs_by_table = defaultdict(list)
    for i, job in enumerate(results):
        jobs_by_table[tasks[i].table].append(job)
    bytes_processed = rows_deleted = 0
    for table, jobs in jobs_by_table.items():
        table_bytes_processed = sum(job.total_bytes_processed or 0 for job in jobs)
        bytes_processed += table_bytes_processed
        table_id = sql_table_id(table)
        if args.dry_run:
            logging.info(f"Would scan {table_bytes_processed} bytes from {table_id}")
        else:
            table_rows_deleted = sum(job.num_dml_affected_rows or 0 for job in jobs)
            rows_deleted += table_rows_deleted
            logging.info(
                f"Scanned {table_bytes_processed} bytes and "
                f"deleted {table_rows_deleted} rows from {table_id}"
            )
    if args.dry_run:
        logging.info(f"Would scan {bytes_processed} in total")
    else:
        logging.info(
            f"Scanned {bytes_processed} and deleted {rows_deleted} rows in total"
        )


if __name__ == "__main__":
    warnings.filterwarnings("ignore", module="google.auth._default")
    main()



import os

import pytest
from click.testing import CliRunner

from bigquery_etl.cli.format import format as sql_format


class TestFormat:
    @pytest.fixture
    def runner(self):
        return CliRunner()

    def test_format_invalid_path(self, runner):
        result = runner.invoke(sql_format, ["not-existing-path.sql"])
        assert result.exit_code == 1

    def test_format(self, runner):
        with runner.isolated_filesystem():
            with open("foo.sql", "w") as f:
                f.write("SELECT 1 FROM test")

            result = runner.invoke(sql_format, ["foo.sql"])
            assert "1 file reformatted." in result.output
            assert result.exit_code == 0

            os.mkdir("test")
            with open("test/foo.sql", "w") as f:
                f.write("SELECT 1 FROM test")
            with open("test/bar.sql", "w") as f:
                f.write("SELECT 1 FROM test")

            result = runner.invoke(sql_format, ["test"])
            assert "2 files reformatted." in result.output
            assert result.exit_code == 0
